#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod
from typing import Any, Dict, Literal, Optional

from argilla.client.feedback.schemas.enums import FieldTypes
from argilla.client.feedback.schemas.validators import title_must_have_value
from argilla.pydantic_v1 import BaseModel, Extra, Field, validator


class FieldSchema(BaseModel, ABC):
    """Base schema for the `FeedbackDataset` fields.

    Args:
        name: The name of the field. This is the only required field.
        title: The title of the field. If not provided, it will be capitalized from
            the `name` field. And its what will be shown in the UI.
        required: Whether the field is required or not. Defaults to True. Note that at
            least one field must be required.
        type: The type of the field. Defaults to None, and ideally it should be defined
            in the class inheriting from this one to be able to use a discriminated union
            based on the `type` field.

    Disclaimer:
        You should not use this class directly, but instead use the classes that inherit
        from this one, as they will have the `type` field already defined, and ensured
        to be supported by Argilla.
    """

    name: str = Field(..., regex=r"^(?=.*[a-z0-9])[a-z0-9_-]+$")
    title: Optional[str] = None
    required: bool = True
    type: Optional[FieldTypes] = Field(..., allow_mutation=False)

    _title_must_have_value = validator("title", always=True, allow_reuse=True)(title_must_have_value)

    class Config:
        validate_assignment = True
        extra = Extra.forbid
        exclude = {"type"}

    @property
    @abstractmethod
    def server_settings(self) -> Dict[str, Any]:
        """Abstract property that should be implemented by the classes that inherit from
        this one, and that will be used to create the `FeedbackDataset` in Argilla.
        """
        ...

    def to_server_payload(self) -> Dict[str, Any]:
        """Method that will be used to create the payload that will be sent to Argilla
        to create a field in the `FeedbackDataset`.
        """
        return {
            "name": self.name,
            "title": self.title,
            "required": self.required,
            "settings": self.server_settings,
        }


class TextField(FieldSchema):
    """Schema for the `FeedbackDataset` text fields, which are the ones that will
    require a text to be defined as part of the record.

    Args:
        type: The type of the field. Defaults to 'text' and cannot/shouldn't be
            modified.
        use_markdown: Whether the field should be rendered using markdown or not.
            Defaults to False.

    Examples:
        >>> from argilla.client.feedback.schemas.fields import TextField
        >>> TextField(name="text_field", title="Text Field")
    """

    type: Literal[FieldTypes.text] = Field(FieldTypes.text.value, allow_mutation=False)
    use_markdown: bool = False

    @property
    def server_settings(self) -> Dict[str, Any]:
        return {
            "type": self.type,
            "use_markdown": self.use_markdown,
        }



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import typer

app = typer.Typer(invoke_without_command=True)


@app.callback(help="Check the current user on the Argilla Server")
def whoami() -> None:
    from rich.markdown import Markdown

    from argilla.cli.callback import init_callback
    from argilla.cli.rich import echo_in_panel
    from argilla.client.singleton import active_client

    init_callback()
    user = active_client()._user

    echo_in_panel(
        Markdown(
            f"- **Username**: {user.username}\n"
            f"- **Role**: {user.role}\n"
            f"- **First name**: {user.first_name}\n"
            f"- **Last name**: {user.last_name}\n"
            f"- **API Key**: {user.api_key}\n"
            f"- **Workspaces**: {user.workspaces}"
        ),
        title="Current User",
        title_align="left",
    )


if __name__ == "__main__":
    app()



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import asyncio
import warnings
from asyncio import Future
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Tuple, Union

from argilla.client.client import Argilla
from argilla.client.datasets import Dataset
from argilla.client.enums import DatasetType
from argilla.client.feedback.dataset.local.dataset import FeedbackDataset
from argilla.client.models import BulkResponse, Record  # TODO Remove TextGenerationRecord
from argilla.client.sdk.commons import errors
from argilla.client.sdk.datasets.models import Dataset as DatasetModel
from argilla.client.sdk.v1.workspaces.models import WorkspaceModel
from argilla.client.singleton import ArgillaSingleton

if TYPE_CHECKING:
    from argilla.client.feedback.dataset.remote.dataset import RemoteFeedbackDataset

Api = Argilla  # Backward compatibility


def log(
    records: Union[Record, Iterable[Record], Dataset],
    name: str,
    workspace: Optional[str] = None,
    tags: Optional[Dict[str, str]] = None,
    metadata: Optional[Dict[str, Any]] = None,
    batch_size: int = 100,
    verbose: bool = True,
    background: bool = False,
    chunk_size: Optional[int] = None,
    num_threads: int = 0,
    max_retries: int = 3,
) -> Union[BulkResponse, Future]:
    """Logs Records to argilla.

    The logging happens asynchronously in a background thread.

    Args:
        records: The record, an iterable of records, or a dataset to log.
        name: The dataset name.
        workspace: The workspace to which records will be logged/loaded. If `None` (default) and the
            env variable ``ARGILLA_WORKSPACE`` is not set, it will default to the private user workspace.
        tags: A dictionary of tags related to the dataset.
        metadata: A dictionary of extra info for the dataset.
        batch_size: The batch size for a data bulk.
        verbose: If True, shows a progress bar and prints out a quick summary at the end.
        background: If True, we will NOT wait for the logging process to finish and return an ``asyncio.Future``
            object. You probably want to set ``verbose`` to False in that case.
        chunk_size: DEPRECATED! Use `batch_size` instead.
        num_threads: If > 0, will use num_thread separate number threads to batches, sending data concurrently.
                Default to `0`, which means no threading at all.
        max_retries: Number of retries when logging a batch of records if a `httpx.TransportError` occurs.
                Default `3`.

    Returns:
        Summary of the response from the REST API.
        If the ``background`` argument is set to True, an ``asyncio.Future`` will be returned instead.

    Examples:
        >>> import argilla as rg
        >>> record = rg.TextClassificationRecord(
        ...     text="my first argilla example",
        ...     prediction=[('spam', 0.8), ('ham', 0.2)]
        ... )
        >>> rg.log(record, name="example-dataset")
        1 records logged to http://localhost:6900/datasets/argilla/example-dataset
        BulkResponse(dataset='example-dataset', processed=1, failed=0)
        >>>
        >>> # Logging records in the background
        >>> rg.log(record, name="example-dataset", background=True, verbose=False)
        <Future at 0x7f675a1fffa0 state=pending>
    """
    return ArgillaSingleton.get().log(
        records=records,
        name=name,
        workspace=workspace,
        tags=tags,
        metadata=metadata,
        batch_size=batch_size,
        verbose=verbose,
        background=background,
        chunk_size=chunk_size,
        num_threads=num_threads,
        max_retries=max_retries,
    )


async def log_async(
    records: Union[Record, Iterable[Record], Dataset],
    name: str,
    workspace: Optional[str] = None,
    tags: Optional[Dict[str, str]] = None,
    metadata: Optional[Dict[str, Any]] = None,
    batch_size: int = 100,
    verbose: bool = True,
    chunk_size: Optional[int] = None,
) -> BulkResponse:
    """Logs Records to argilla with asyncio.

    Args:
        records: The record, an iterable of records, or a dataset to log.
        name: The dataset name.
        workspace: The workspace to which records will be logged/loaded. If `None` (default) and the
            env variable ``ARGILLA_WORKSPACE`` is not set, it will default to the private user workspace.
        tags: A dictionary of tags related to the dataset.
        metadata: A dictionary of extra info for the dataset.
        batch_size: The batch size for a data bulk.
        verbose: If True, shows a progress bar and prints out a quick summary at the end.
        chunk_size: DEPRECATED! Use `batch_size` instead.

    Returns:
        Summary of the response from the REST API

    Examples:
        >>> # Log asynchronously from your notebook
        >>> import asyncio
        >>> import argilla as rg
        >>> from argilla.utils import setup_loop_in_thread
        >>> loop, _ = setup_loop_in_thread()
        >>> future_response = asyncio.run_coroutine_threadsafe(
        ...     rg.log_async(my_records, dataset_name), loop
        ... )
    """

    warnings.warn(
        (
            "`log_async` is deprecated and will be removed in next release. "
            "Please, use `log` with `background=True` instead"
        ),
        DeprecationWarning,
    )

    future = ArgillaSingleton.get().log(
        records=records,
        name=name,
        workspace=workspace,
        tags=tags,
        metadata=metadata,
        batch_size=batch_size,
        verbose=verbose,
        chunk_size=chunk_size,
        background=True,
    )

    return await asyncio.wrap_future(future)


def load(
    name: str,
    workspace: Optional[str] = None,
    query: Optional[str] = None,
    vector: Optional[Tuple[str, List[float]]] = None,
    ids: Optional[List[Union[str, int]]] = None,
    limit: Optional[int] = None,
    sort: Optional[List[Tuple[str, str]]] = None,
    id_from: Optional[str] = None,
    batch_size: int = 250,
    include_vectors: bool = True,
    include_metrics: bool = True,
    as_pandas: Optional[bool] = None,
) -> Union[Dataset, "RemoteFeedbackDataset"]:
    """Loads a argilla dataset.

    Args:
        name: The dataset name.
        workspace: The workspace to which records will be logged/loaded. If `None` (default) and the
            env variable ``ARGILLA_WORKSPACE`` is not set, it will default to the private user workspace.
        query: An ElasticSearch query with the `query string
            syntax <https://docs.argilla.io/en/latest/practical_guides/filter_dataset.html>`_
        vector: Vector configuration for a semantic search
        ids: If provided, load dataset records with given ids.
        limit: The number of records to retrieve.
        sort: The fields on which to sort [(<field_name>, 'asc|decs')].
        id_from: If provided, starts gathering the records starting from that Record.
            As the Records returned with the load method are sorted by ID, `id_from`
            can be used to load using batches.
        batch_size: If provided, load `batch_size` samples per request. A lower batch
            size may help avoid timeouts.
        include_vectors: When set to `False`, indicates that records will be retrieved excluding their vectors,
            if any. By default, this parameter is set to `True`, meaning that vectors will be included.
        include_metrics: When set to `False`, indicates that records will be retrieved excluding their metrics.
            By default, this parameter is set to `True`, meaning that metrics will be included.
        as_pandas: DEPRECATED! To get a pandas DataFrame do
            ``rg.load('my_dataset').to_pandas()``.

    Returns:
        A argilla dataset.

    Examples:
        **Basic Loading: load the samples sorted by their ID**

        >>> import argilla as rg
        >>> dataset = rg.load(name="example-dataset")

        **Iterate over a large dataset:**
            When dealing with a large dataset you might want to load it in batches to optimize memory consumption
            and avoid network timeouts. To that end, a simple batch-iteration over the whole database can be done
            employing the `from_id` parameter. This parameter will act as a delimiter, retrieving the N items after
            the given id, where N is determined by the `limit` parameter. **NOTE** If
            no `limit` is given the whole dataset after that ID will be retrieved.

        >>> import argilla as rg
        >>> dataset_batch_1 = rg.load(name="example-dataset", limit=1000)
        >>> dataset_batch_2 = rg.load(name="example-dataset", limit=1000, id_from=dataset_batch_1[-1].id)

    """
    argilla = ArgillaSingleton.get()
    try:
        return argilla.load(
            name=name,
            workspace=workspace,
            query=query,
            vector=vector,
            ids=ids,
            limit=limit,
            sort=sort,
            id_from=id_from,
            batch_size=batch_size,
            include_metrics=include_metrics,
            include_vectors=include_vectors,
            as_pandas=as_pandas,
        )
    except errors.ArApiResponseError as e:
        workspace = workspace or argilla.get_workspace()
        try:
            dataset = FeedbackDataset.from_argilla(name=name, workspace=workspace)
        except ValueError:
            raise e

        warnings.warn(
            "Loaded dataset is a `FeedbackDataset`. It's recommended to use "
            f"`rg.FeedbackDataset.from_argilla(name='{name}', workspace='{workspace}')` instead",
            UserWarning,
            stacklevel=2,
        )

        return dataset


def copy(dataset: str, name_of_copy: str, workspace: Optional[str] = None) -> None:
    """
    Creates a copy of a dataset including its tags and metadata

    Args:
        dataset: Name of the source dataset
        name_of_copy: Name of the copied dataset
        workspace: If provided, dataset will be copied to that workspace

    Examples:
        >>> import argilla as rg
        >>> rg.copy("my_dataset", name_of_copy="new_dataset")
        >>> rg.load("new_dataset")
    """
    ArgillaSingleton.get().copy(
        dataset=dataset,
        name_of_copy=name_of_copy,
        workspace=workspace,
    )


def delete(name: str, workspace: Optional[str] = None) -> None:
    """
    Deletes an Argilla dataset from the server. It can be used with both `Dataset` and `FeedbackDataset`, although
    for the latter it's recommended to use `rg.FeedbackDataset.delete` instead.

    Args:
        name: The name of the dataset to delete.
        workspace: The workspace to which the dataset belongs. If `None` (default) and the env variable
            ``ARGILLA_WORKSPACE`` is not set, it will default to the private user workspace.

    Raises:
        ValueError: If no dataset is found with the given name and workspace.
        PermissionError: If the dataset that's being deleted is a `FeedbackDataset` and the user doesn't have enough
            permissions to delete it.
        RuntimeError: If the dataset that's being deleted is a `FeedbackDataset` and some kind of error occurs during
            the deletion process.

    Examples:
        >>> import argilla as rg
        >>> rg.delete(name="example-dataset")
    """
    argilla = ArgillaSingleton.get()
    workspace = workspace or argilla.get_workspace()
    try:
        # `delete` method is always successful, even if the dataset does not exist and that's why we need the extra
        # call to `get_dataset` to check if the dataset exists. If it doesn't exist, then we try to delete a `FeedbackDataset`.
        argilla.get_dataset(name=name, workspace=workspace)
        argilla.delete(name=name, workspace=workspace)
    except errors.ArApiResponseError as e:
        try:
            dataset = FeedbackDataset.from_argilla(name=name, workspace=workspace)
        except ValueError:
            raise e

        dataset.delete()

        warnings.warn(
            "Removed dataset was a `FeedbackDataset`. It's recommended to use "
            f"`rg.FeedbackDataset.from_argilla(name='{name}', workspace='{workspace}').delete()` instead.",
            UserWarning,
            stacklevel=2,
        )


def delete_records(
    name: str,
    workspace: Optional[str] = None,
    query: Optional[str] = None,
    ids: Optional[List[Union[str, int]]] = None,
    discard_only: bool = False,
    discard_when_forbidden: bool = True,
) -> Tuple[int, int]:
    """Delete records from a argilla dataset.

    Args:
        name: The dataset name.
        workspace: The workspace to which records will be logged/loaded. If `None` (default) and the
            env variable ``ARGILLA_WORKSPACE`` is not set, it will default to the private user workspace.
        query: An ElasticSearch query with the `query string syntax
            <https://docs.argilla.io/en/latest/practical_guides/filter_dataset.html>`_
        ids: If provided, deletes dataset records with given ids.
        discard_only: If `True`, matched records won't be deleted. Instead, they will be marked as `Discarded`
        discard_when_forbidden: Only super-user or dataset creator can delete records from a dataset.
            So, running "hard" deletion for other users will raise an `ForbiddenApiError` error.
            If this parameter is `True`, the client API will automatically try to mark as ``Discarded``
            records instead. Default, `True`

    Returns:
        The total of matched records and real number of processed errors. These numbers could not
        be the same if some data conflicts are found during operations (some matched records change during
        deletion).

    Examples:
        >>> ## Delete by id
        >>> import argilla as rg
        >>> rg.delete_records(name="example-dataset", ids=[1,3,5])
        >>> ## Discard records by query
        >>> import argilla as rg
        >>> rg.delete_records(name="example-dataset", query="metadata.code=33", discard_only=True)
    """
    return ArgillaSingleton.get().delete_records(
        name=name,
        workspace=workspace,
        query=query,
        ids=ids,
        discard_only=discard_only,
        discard_when_forbidden=discard_when_forbidden,
    )


def set_workspace(workspace: str) -> None:
    """Sets the active workspace.

    Args:
        workspace: The new workspace
    """
    ArgillaSingleton.get().set_workspace(workspace=workspace)


def get_workspace() -> str:
    """Returns the name of the active workspace.

    Returns:
        The name of the active workspace as a string.
    """
    return ArgillaSingleton.get().get_workspace()


def list_workspaces() -> List[WorkspaceModel]:
    """Lists all the available workspaces for the current user.

    Returns:
        A list of `WorkspaceModel` objects, containing the workspace
        attributes: name, id, created_at, and updated_at.
    """
    warnings.warn(
        "`Workspace.list` is recommended over `list_workspaces`, since you can easily"
        " access the workspaces as a list of `Workspace` objects with their attributes"
        " and methods.",
        UserWarning,
        stacklevel=1,
    )
    return ArgillaSingleton.get().list_workspaces()


def list_datasets(
    workspace: Optional[str] = None, type: Optional[Union[str, DatasetType]] = None
) -> List[Union[DatasetModel, "RemoteFeedbackDataset"]]:
    """Lists all the available datasets for the current user in Argilla.

    Args:
        workspace: If provided, list datasets from that workspace only. Note that
            the workspace must exist in advance, otherwise a HTTP 400 error will be
            raised.

    Returns:
        A list of `DatasetModel` objects, containing the dataset
        attributes: tags, metadata, name, id, task, owner, workspace, created_at,
        and last_updated.
    """
    if type is not None:
        type = type if isinstance(type, DatasetType) else DatasetType(type)

    old_datasets = []
    if type is None or type == DatasetType.other:
        old_datasets = ArgillaSingleton.get().list_datasets(workspace=workspace)

    datasets = []
    if type is None or type == DatasetType.feedback:
        datasets = FeedbackDataset.list(workspace=workspace)

    return old_datasets + datasets



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import random
import warnings
from collections import defaultdict
from typing import Any, Dict, List, Union

from rich.progress import Progress

from argilla.client.users import User
from argilla.client.workspaces import Workspace


def check_user(user_to_check: Union[str, User]) -> User:
    """
    Helper function to check if the input is a User object. If it's a string, it attempts to retrieve the User object.
    If the User does not exist, it creates a new User with a default password and role.

    Args:
        user_to_check: a user object or a string that represents a username

    Returns:
        The User object corresponding to the input.
    """
    if isinstance(user_to_check, User):
        user = user_to_check
    else:
        try:
            user = User.from_name(user_to_check)
        except ValueError:
            user = User.create(username=user_to_check, first_name=user_to_check, password="12345678", role="annotator")
            warnings.warn(
                f"The user {user.username} was created with a default password. We recommend you to change it for security reasons.",
                UserWarning,
            )
    return user


def check_workspace(workspace_to_check: str) -> Workspace:
    """
    Helper function to check if the workspace exists. If it does not exist, it creates a new one.

    Args:
        workspace_to_check: a workspace string name

    Returns:
        The Workspace object corresponding to the input.
    """
    try:
        workspace = Workspace.from_name(workspace_to_check)
    except:
        workspace = Workspace.create(workspace_to_check)
    return workspace


def assign_records_to_groups(
    groups: Dict[str, List[Any]], records: List[Any], overlap: int, shuffle: bool = True
) -> Dict[str, Dict[str, Any]]:
    """
    Assign records to predefined groups with controlled overlap (for the groups) and optional shuffle. All members of the same group will annotate the same records.

    Args:
        groups: A dictionary where keys are group names and values are lists of users names or objects.
        records: A list of records to be assigned.
        overlap: The number of times each record is assigned to consecutive groups (0 for no overlap).
        shuffle: If True, shuffle the records before assignment. Defaults to True.

    Returns:
        A dictionary where each key is a group and its value is another dictionary, which maps usernames to their respective assigned records.

    Raises:
        ValueError: If `overlap` is higher than the number of groups or negative.
    """
    if overlap < 0 or overlap >= len(groups.keys()):
        raise ValueError("Overlap must be less than the number of groups and must not be negative.")

    if len(records) < len(groups.keys()):
        warnings.warn(
            "The number of groups is higher than the number of records. Some users will not be assigned any records.",
            UserWarning,
        )

    if shuffle:
        random.shuffle(records)

    assignments = {}
    assignments_grouped = {}
    group_names = list(groups.keys())
    num_groups = len(group_names)
    overlap = 1 if overlap == 0 else overlap

    group_records = defaultdict(list)
    with Progress() as progress:
        task = progress.add_task("[green]Processing records...", total=len(records))

        for idx, record in enumerate(records):
            for offset in range(overlap):
                group_index = (idx + offset) % num_groups
                group_name = group_names[group_index]
                group_records[group_name].append(record)

            progress.update(task, advance=1)

    for group, users in groups.items():
        users = [check_user(user) for user in users]
        for user in users:
            assignments[user] = group_records[group]

        assignments_grouped[group] = {user.username: assignments.get(user, []) for user in users}

    return assignments_grouped


def assign_records_to_individuals(
    users: List[Any], records: List[Any], overlap: int, shuffle: bool = True
) -> Dict[str, List[Any]]:
    """
    Assign records to users with controlled overlap and optional shuffle.

    Args:
        users: A list of user objects, each with a 'username' attribute.
        records: A list of record objects to be assigned to users.
        overlap: The number of times each record is assigned to consecutive users (0 for no overlap).
        shuffle: If True, the records list will be shuffled before assignment. Defaults to True.

    Returns:
        A dictionary where keys are usernames and values are lists of assigned records.

    Raises:
        ValueError: If `overlap` is higher than the number of users or negative.
    """
    if overlap < 0 or overlap >= len(users):
        raise ValueError("Overlap must be less than the number of users and must not be negative.")

    if len(records) < len(users):
        warnings.warn(
            "The number of users is higher than the number of records. Some users will not be assigned any records.",
            UserWarning,
        )

    if shuffle:
        random.shuffle(records)

    users = [check_user(user) for user in users]
    assignments = {user.username: [] for user in users}

    num_users = len(users)
    overlap = 1 if overlap == 0 else overlap

    with Progress() as progress:
        task = progress.add_task("[green]Processing records...", total=len(records))

        for idx, record in enumerate(records):
            for offset in range(overlap):
                user_index = (idx + offset) % num_users
                user = users[user_index].username
                assignments[user].append(record)

            progress.update(task, advance=1)

    return assignments


def assign_records(
    users: Union[Dict[str, List[Any]], List[Any]], records: List[Any], overlap: int, shuffle: bool = True
) -> Union[Dict[str, List[Any]], Dict[str, Dict[str, Any]]]:
    """
    Assign records to either groups or individuals, with controlled overlap and optional shuffle.

    Args:
        users: Either a dictionary of groups or a list of individual user objects.
        records: A list of record objects to be assigned.
        overlap: The number of times each record is assigned to consecutive users or groups (0 for no overlap).
        shuffle: If True, the records list will be shuffled before assignment. Defaults to True.

    Returns:
        A dictionary where each key is a group and its value is another dictionary, which maps usernames to their respective assigned records.
        Or a dictionary where keys are usernames and values are lists of assigned records.

    Examples:
        >>> from argilla.client.feedback.utils import assign_records
        >>> individual_assignments = assign_records([user1, user2, user3], records, 0, False)
        >>> group_assignments = assign_records({group1: [user1, user2], group2: [user3]}, records, 1, False)

    """
    if isinstance(users, dict):
        return assign_records_to_groups(users, records, overlap, shuffle)
    elif isinstance(users, list):
        return assign_records_to_individuals(users, records, overlap, shuffle)


def assign_workspaces(
    assignments: Union[Dict[str, List[Any]], Dict[str, Dict[str, Any]]], workspace_type: str
) -> Dict[str, List[Any]]:
    """
    Assign workspaces (and create them if needed) to either groups or individuals.

    Args:
        assignments: Either a dictionary of groups or a dictionary of users.
        workspace_type: Either 'group' (each group in a workspace), 'group_personal' (each member in a workspace) or 'individual' (each person in a workspace).

    Returns:
        A dictionary where each key is a workspace name and its value is a list of user names.

    Examples:
        >>> from argilla.client.feedback.utils import assign_workspaces
        >>> wk_assignments = assign_workspaces(group_assignments, "group")
        >>> wk_assignments = assign_workspaces(group_assignments, "group_personal")
        >>> wk_assignments = assign_workspaces(individual_assignments, "individual")

    """
    wk_assignments = {}

    for group, users in assignments.items():
        if workspace_type == "group":
            workspace_name = group
            user_ids = [check_user(user).id for user in users.keys()]

        elif workspace_type == "group_personal":
            for user in users.keys():
                workspace_name = user
                user_ids = [check_user(user).id]
                workspace = check_workspace(workspace_name)

                for user_id in user_ids:
                    try:
                        workspace.add_user(user_id)
                    except:
                        pass

                wk_assignments[workspace_name] = [User.from_id(user.id).username for user in workspace.users]

            continue

        elif workspace_type == "individual":
            workspace_name = group
            user_ids = [check_user(group).id]

        workspace = check_workspace(workspace_name)

        for user_id in user_ids:
            try:
                workspace.add_user(user_id)
            except:
                pass

        wk_assignments[workspace_name] = [User.from_id(user.id).username for user in workspace.users]

    return wk_assignments



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from argilla.client.feedback.utils.assignment import (
    assign_records,
    assign_records_to_groups,
    assign_records_to_individuals,
    assign_workspaces,
    check_user,
    check_workspace,
)
from argilla.client.feedback.utils.html_utils import (
    audio_to_html,
    create_token_highlights,
    get_file_data,
    image_to_html,
    is_valid_dimension,
    media_to_html,
    pdf_to_html,
    validate_media_type,
    video_to_html,
)

__all__ = [
    "audio_to_html",
    "video_to_html",
    "image_to_html",
    "pdf_to_html",
    "create_token_highlights",
    "assign_records",
    "assign_workspaces",
]



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from typing import TYPE_CHECKING, List, Optional, Tuple, Union

from argilla.client.feedback.training.base import ArgillaTrainerSkeleton
from argilla.client.feedback.training.schemas.base import TrainingTaskForSentenceSimilarity
from argilla.training.utils import filter_allowed_args, get_default_args
from argilla.utils.dependency import require_dependencies

if TYPE_CHECKING:
    from argilla.client.feedback.dataset import FeedbackDataset
    from argilla.client.feedback.integrations.huggingface.model_card import SentenceTransformerCardData


class ArgillaSentenceTransformersTrainer(ArgillaTrainerSkeleton):
    _logger = logging.getLogger("ArgillaSentenceTransformersTrainer")
    _logger.setLevel(logging.INFO)

    def __init__(
        self,
        dataset: "FeedbackDataset",
        task: TrainingTaskForSentenceSimilarity,
        prepared_data=None,
        model: str = None,
        seed: int = None,
        train_size: Optional[float] = 1,
        cross_encoder: bool = False,
    ) -> None:
        require_dependencies("sentence-transformers")

        super().__init__(
            dataset=dataset, task=task, prepared_data=prepared_data, model=model, train_size=train_size, seed=seed
        )

        # The prepared_data is lost in the TrainerSkeleton, is this intended?
        self._prepared_data = prepared_data
        self._cross_encoder = cross_encoder

        if model is None:
            if not self._cross_encoder:
                self._model = "sentence-transformers/all-MiniLM-L6-v2"
            else:
                self._model = "cross-encoder/ms-marco-MiniLM-L-6-v2"

            self._logger.info(
                f"No model was selected, using pre-defined `{'Cross-Encoder' if self._cross_encoder else 'Bi-Encoder'}` with `{self._model}`."
            )

        else:
            self._model = model

        if isinstance(self._prepared_data, tuple):
            self._train_dataset = self._prepared_data[0]
            self._eval_dataset = self._prepared_data[1]
        else:
            self._train_dataset = self._prepared_data
            self._eval_dataset = None

        self._loss_cls = None
        self.init_training_args()

    def init_training_args(self) -> None:
        """
        Initializes the training arguments.
        """
        self.model_kwargs = {}

        # Use a sample of the dataset to do some checking
        sample = self._train_dataset[0]

        if self._cross_encoder:
            # Cross encoders don't support training with triplets
            if (n := len(sample.texts)) > 2:
                raise ValueError(
                    f"Cross-encoders don't support training with triplets, the dataset has `{n}` sentences per sample."
                )

            from sentence_transformers import CrossEncoder

            self._trainer_cls = CrossEncoder
        else:
            from sentence_transformers import SentenceTransformer

            self._trainer_cls = SentenceTransformer

        self.model_kwargs = get_default_args(self._trainer_cls.__init__)
        self.trainer_kwargs = get_default_args(self._trainer_cls.fit)
        # These arguments don't exactly fit in either of the previous kwargs.
        # We store here arguments that can be passed to the DataLoader or
        # related to the dataset types defined in sentence-transformers.
        self.data_kwargs = {}
        self.data_kwargs["batch_size"] = 32
        self.data_kwargs["dataset_type"] = None

        # For a guide on the selection of the loss function:
        # https://huggingface.co/blog/how-to-train-sentence-transformers#loss-functions-for-training-a-sentence-transformers-model
        self._set_loss_cls(sample.label, len(sample.texts))
        self._logger.warning(f"`loss_cls` parameter set as default `{self._loss_cls}`.")

        if self._eval_dataset:
            self._set_evaluator(label=sample.label, number_of_sentences=len(sample.texts))
            self._logger.warning(f"`evaluator` parameter set to `{self.trainer_kwargs['evaluator']}` by default.")

    def _set_loss_cls(self, label: Optional[Union[int, float]], number_of_sentences: int) -> None:
        """
        Helper method to choose a loss class for the model based on the type
        of model, whether it has label or not and the number of sentences per example.

        Args:
            label: The label type, None should be passed if there is no label.
            number_of_sentences: number of sentences (two or three normally).
        """
        if label:
            if number_of_sentences == 2:
                if isinstance(label, int):
                    from sentence_transformers.losses import ContrastiveLoss

                    self._loss_cls = ContrastiveLoss
                else:
                    from sentence_transformers.losses import CosineSimilarityLoss

                    self._loss_cls = CosineSimilarityLoss

            else:
                if isinstance(label, int):
                    from sentence_transformers.losses import BatchHardTripletLoss

                    self._loss_cls = BatchHardTripletLoss
                else:
                    from sentence_transformers.losses import BatchAllTripletLoss

                    self._loss_cls = BatchAllTripletLoss
                    from sentence_transformers.datasets import SentenceLabelDataset

                    self.data_kwargs["dataset_type"] = SentenceLabelDataset

        else:
            if number_of_sentences == 3:
                from sentence_transformers.losses import TripletLoss

                self._loss_cls = TripletLoss
            else:
                from sentence_transformers.losses import MultipleNegativesRankingLoss

                self._loss_cls = MultipleNegativesRankingLoss

    def _set_evaluator(
        self, label: Optional[Union[int, float]] = None, number_of_sentences: int = None, evaluator_cls=None
    ) -> None:
        """
        Helper method to choose an evaluator for the model based on the type
        of model, whether it has label or not and the number of sentences per example.
        If neither label or number_of_sentences is set, it's assumed that evaluator_cls
        is passed (via update_config).

        Args:
            label: The label type, None should be passed if there is no label.
            number_of_sentences: Number of sentences (two or three normally).
            evaluator_cls: Class to set without "best guess".
        """
        if evaluator_cls:
            # When the user wants to set an evaluator on its own
            if self._eval_dataset:
                self._logger.warning("No `eval_dataset` found, `evaluator` cannot be set.")
            else:
                try:
                    self.trainer_kwargs["evaluator"] = evaluator_cls.from_input_examples(self._eval_dataset)
                except AttributeError:
                    # For the moment it's easier (and covers most of the cases?) to allow only
                    # a subset of the evaluators defined in sentence-transformers, inform
                    # the user the one selected is not allowed.
                    self.trainer_kwargs["evaluator"] = None
                    self._logger.warning(
                        "Currently only developed evaluators that implement `cls.from_input_examples` can be set."
                    )
        else:
            if label:
                if number_of_sentences == 2:
                    if isinstance(label, int):
                        # This should work, but for some reason in the tests with default values
                        # the classes cannot be instantiated, need some time to review.
                        # if self._cross_encoder:
                        #     from sentence_transformers.cross_encoder.evaluation import CEBinaryClassificationEvaluator
                        #     self.trainer_kwargs["evaluator"] = CEBinaryClassificationEvaluator.from_input_examples(self._eval_dataset)
                        # else:
                        #     from sentence_transformers.evaluation import BinaryClassificationEvaluator
                        #     self.trainer_kwargs["evaluator"] = BinaryClassificationEvaluator.from_input_examples(self._eval_dataset)
                        self.trainer_kwargs["evaluator"] = None

                    else:
                        if self._cross_encoder:
                            from sentence_transformers.cross_encoder.evaluation import CECorrelationEvaluator

                            self.trainer_kwargs["evaluator"] = CECorrelationEvaluator.from_input_examples(
                                self._eval_dataset
                            )
                        else:
                            from sentence_transformers.evaluation import EmbeddingSimilarityEvaluator

                            self.trainer_kwargs["evaluator"] = EmbeddingSimilarityEvaluator.from_input_examples(
                                self._eval_dataset
                            )

                else:
                    from sentence_transformers.evaluation import TripletEvaluator

                    self.trainer_kwargs["evaluator"] = TripletEvaluator.from_input_examples(self._eval_dataset)
            else:
                if number_of_sentences == 3:
                    from sentence_transformers.evaluation import TripletEvaluator

                    self.trainer_kwargs["evaluator"] = TripletEvaluator.from_input_examples(self._eval_dataset)
                else:
                    # Hard to choose one here
                    self.trainer_kwargs["evaluator"] = None

    def init_model(self) -> None:
        """
        Initializes a model.
        """
        if "model_name_or_path" in self.model_kwargs:
            self.model_kwargs.pop("model_name_or_path")
        if "model_name" in self.model_kwargs:
            self.model_kwargs.pop("model_name")

        if "train_objectives" in self.trainer_kwargs:
            self.trainer_kwargs.pop("train_objectives")
        if "train_dataloader" in self.trainer_kwargs:
            self.trainer_kwargs.pop("train_dataloader")

        self._trainer = self._trainer_cls(self._model, **self.model_kwargs)

    def update_config(self, **kwargs) -> None:
        """
        Updates the configuration of the trainer, but the parameters depend on the trainer.subclass.
        """
        self.model_kwargs.update(filter_allowed_args(self._trainer_cls.__init__, **kwargs))
        self.trainer_kwargs.update(filter_allowed_args(self._trainer_cls.fit, **kwargs))

        if "batch_size" in kwargs:
            self.data_kwargs["batch_size"] = kwargs.pop("batch_size")

        # These parameters are allowed, but for simplicity
        # they are passed as self._model as positional argument to avoid errors
        # between Cross-Encoder/Bi-encoder.
        if "model_name_or_path" in kwargs:
            self._model = self.model_kwargs.pop("model_name_or_path")
        if "model_name" in kwargs:
            self._model = self.model_kwargs.pop("model_name")

        if "loss_cls" in kwargs:
            self._loss_cls = kwargs.pop("loss_cls")

        if "evaluator" in kwargs:
            self._set_evaluator(evaluator_cls=kwargs.pop("evaluator"))

    def train(self, output_dir: Optional[str] = None) -> None:
        """
        Trains the model.
        """
        from torch.utils.data import DataLoader

        self.init_model()

        if self.data_kwargs["dataset_type"]:
            dataloader = DataLoader(
                dataset=self.data_kwargs["dataset_type"](self._train_dataset), batch_size=self.data_kwargs["batch_size"]
            )
        else:
            dataloader = DataLoader(dataset=self._train_dataset, batch_size=self.data_kwargs["batch_size"])

        train_loss = self._loss_cls(self._trainer)

        if self._cross_encoder:
            train_objectives = dataloader
        else:
            train_objectives = [(dataloader, train_loss)]

        self._trainer.fit(train_objectives, **self.trainer_kwargs)

        if output_dir is not None:
            self.save(output_dir)

    def predict(
        self, text: Union[List[List[str]], Tuple[str, List[str]]], as_argilla_records: bool = False, **kwargs
    ) -> List[float]:
        """
        Predicts the similarity of the sentences.

        Args:
            text: The sentences to obtain the similarity from.
                Allowed inputs are:
                - A list with a single sentence (as a string) and a list of sentences to compare against.
                - A list with pair of sentences.
            as_argilla_records: If True, the prediction will be returned as an Argilla record. If
                False, the prediction will be returned as a string. Defaults to True

        Returns:
            A list of predicted similarities.
        """
        if isinstance(text[0], str):
            sentences = [(text[0], sentence) for sentence in text[1]]
        else:
            sentences = text

        if self._cross_encoder:
            prediction = list(self._trainer.predict(sentences, **kwargs))

        else:
            from sentence_transformers.util import cos_sim

            # Under this model we could potentially obtain the similarity of all vs all
            # the input sentences, but maybe is more intuitive to return only the similarities
            # between pairs as we do with the CrossEncoder predictions.
            sources, targets = zip(*sentences)
            embeds_source = self._trainer.encode(sources, **kwargs)
            embeds_target = self._trainer.encode(targets, **kwargs)
            prediction = list(cos_sim(embeds_source, embeds_target).numpy()[0])

        if as_argilla_records:
            raise NotImplementedError("To be done")

        return prediction

    def save(self, output_dir: str) -> None:
        """
        Saves the model to the specified path.
        """
        if self._cross_encoder:
            self._trainer.save(output_dir)
        else:
            # For the moment just save the in the simplest form, creating the model card or the
            # dataset for example should be done taking the extra information from the argilla
            # dataset instead of the defaults
            self._trainer.save(output_dir, model_name=None, create_model_card=False, train_datasets=None)

    def get_model_card_data(self, **card_data_kwargs) -> "SentenceTransformerCardData":
        """
        Generate the card data to be used for the `ArgillaModelCard`.

        Args:
            card_data_kwargs: Extra arguments provided by the user when creating the `ArgillaTrainer`.

        Returns:
            SentenceTransformerCardData: Container for the data to be written on the `ArgillaModelCard`.
        """
        from argilla.client.feedback.integrations.huggingface.model_card import SentenceTransformerCardData

        return SentenceTransformerCardData(
            model_id=self._model,
            task=self._task,
            framework_kwargs={"cross_encoder": self._cross_encoder},
            update_config_kwargs={**self.trainer_kwargs, **self.model_kwargs, **self.data_kwargs},
            trainer_cls=self._trainer_cls,
            **card_data_kwargs,
        )

    def push_to_huggingface(self, repo_id: str, **kwargs) -> None:
        """Uploads the model to [huggingface's model hub](https://huggingface.co/models).

        The full list of parameters can be seen at:
        [sentence-transformer api docs](https://www.sbert.net/docs/package_reference/SentenceTransformer.html#sentence_transformers.SentenceTransformer.save_to_hub).

        Args:
            repo_id:
                The name of the repository you want to push your model and tokenizer to.
                It should contain your organization name when pushing to a given organization.

        Raises:
            NotImplementedError:
                For `CrossEncoder` models, that currently aren't implemented underneath.
        """
        raise NotImplementedError("This method is not implemented for `ArgillaSentenceTransformersTrainer`.")



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime
from uuid import UUID

from argilla.client.feedback.schemas.vector_settings import VectorSettings
from argilla.client.sdk.v1.datasets.models import FeedbackVectorSettingsModel


class RemoteVectorSettings(VectorSettings):
    id: UUID
    inserted_at: datetime
    updated_at: datetime

    def to_local(self) -> VectorSettings:
        return VectorSettings(name=self.name, title=self.title, dimensions=self.dimensions)

    @classmethod
    def from_api(cls, api_model: FeedbackVectorSettingsModel) -> "RemoteVectorSettings":
        return RemoteVectorSettings(
            id=api_model.id,
            name=api_model.name,
            title=api_model.title,
            dimensions=api_model.dimensions,
            inserted_at=api_model.inserted_at,
            updated_at=api_model.updated_at,
        )



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, Dict, List, Optional, Union

import httpx

from argilla.client.sdk._helpers import build_raw_response
from argilla.client.sdk.client import AuthenticatedClient
from argilla.client.sdk.commons.api import build_list_response
from argilla.client.sdk.commons.models import ErrorMessage, HTTPValidationError, Response
from argilla.client.sdk.metrics.models import MetricInfo


def get_dataset_metrics(
    client: AuthenticatedClient, name: str, task: str
) -> Response[Union[List[MetricInfo], ErrorMessage, HTTPValidationError]]:
    url = "{}/api/datasets/{task}/{name}/metrics".format(client.base_url, name=name, task=task)

    response = httpx.get(
        url=url,
        headers=client.get_headers(),
        cookies=client.get_cookies(),
        timeout=client.get_timeout(),
    )

    return build_list_response(response, item_class=MetricInfo)


def compute_metric(
    client: AuthenticatedClient, name: str, task: str, metric: str, query: Optional[str] = None, **query_params
) -> Response[Union[Dict[str, Any], ErrorMessage, HTTPValidationError]]:
    url = "{}/api/datasets/{task}/{name}/metrics/{metric}:summary".format(
        client.base_url, task=task, name=name, metric=metric
    )

    query_params = {k: v for k, v in query_params.items() if v is not None}
    if query_params:
        url += "?" + "&".join([f"{k}={v}" for k, v in query_params.items()])

    if query == "":
        query = None

    response = httpx.post(
        url=url,
        headers=client.get_headers(),
        cookies=client.get_cookies(),
        timeout=client.get_timeout(),
        json={"query_text": query},
    )

    return build_raw_response(response)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import warnings
from typing import TYPE_CHECKING, Dict, List, Optional, Union

from argilla.client.feedback.dataset.base import FeedbackDatasetBase
from argilla.client.feedback.schemas.questions import (
    LabelQuestion,
    MultiLabelQuestion,
    RankingQuestion,
    RatingQuestion,
    TextQuestion,
)
from argilla.client.feedback.unification import (
    LabelQuestionStrategy,
    MultiLabelQuestionStrategy,
    RankingQuestionStrategy,
    RatingQuestionStrategy,
    TextQuestionStrategy,
)

if TYPE_CHECKING:
    from argilla.client.feedback.dataset.local.dataset import FeedbackDataset
    from argilla.client.feedback.metrics.agreement_metrics import AgreementMetricResult
    from argilla.client.feedback.metrics.annotator_metrics import ModelMetricResult


class MetricsMixin:
    """Mixin to add functionality to compute the metrics directly from a `FeedbackDataset`."""

    def compute_model_metrics(
        self,
        metric_names: Union[str, List[str]] = None,
        question_name: Union[
            str, LabelQuestion, MultiLabelQuestion, RatingQuestion, TextQuestion, RankingQuestion
        ] = None,
        strategy: Optional[
            Union[str, LabelQuestionStrategy, MultiLabelQuestion, RatingQuestionStrategy, RankingQuestion]
        ] = None,
    ) -> Union[Dict[str, List["ModelMetricResult"]], "ModelMetricResult", List["ModelMetricResult"]]:
        """Compute metrics for the annotators using the suggestions as the ground truth, and the responses
        as the predicted value, or if a strategy is provided, the same but applied to unified responses.

        The metric interpretation is the same whether the responses are unified or not.

        Args:
            metric_names: Metric name or list of metric names of the metrics, dependent on the question type.
            question_name: Question for which we want to compute the metrics.
            strategy: Unification strategy. If given, will unify the responses of the dataset and compute
                the metrics on the unified responses vs the suggestions instead on a per user level.
                See `unified_responses` method for more information. Defaults to None.

        Note:
            Currently, the following types of questions are supported:
            - For annotator level questions: all the types of questions
            - For unified responses: all the questions except the `TextQuestion`.

        Returns:
            metrics_container: If strategy is provided it will unify the annotations and return
                the metrics for the unified responses. Otherwise, it will return the metrics for
                each annotator as a dict, where the key corresponds to the annotator id and the
                values are a list with the metrics.
        """
        from argilla.client.feedback.metrics.annotator_metrics import ModelMetric, UnifiedModelMetric

        if strategy:
            self.compute_unified_responses(question_name, strategy)
            return UnifiedModelMetric(self, question_name).compute(metric_names)
        else:
            return ModelMetric(self, question_name).compute(metric_names)

    def compute_agreement_metrics(
        self,
        metric_names: Union[str, List[str]] = None,
        question_name: Union[str, LabelQuestion, MultiLabelQuestion, RatingQuestion, RankingQuestion] = None,
    ) -> Union["AgreementMetricResult", List["AgreementMetricResult"]]:
        """Compute agreement or reliability of annotation metrics.

        This metrics can be used to determine the level of agreement across our annotation team,
        or whether the guidelines are clear enough for example.

        Args:
            metric_names: Metric name or list of metric names of the metrics, dependent on the question type.
            question_name: Question for which we want to compute the metrics.

        Note:
            Currently, TextQuestion is not supported.

        Returns:
            metrics_result: Agreement metrics result or a list of metrics results if a list of metric
                names is provided.
        """
        from argilla.client.feedback.metrics.agreement_metrics import AgreementMetric

        return AgreementMetric(self, question_name).compute(metric_names)


class UnificationMixin:
    def unify_responses(
        self: "FeedbackDatasetBase",
        question: Union[str, LabelQuestion, MultiLabelQuestion, RatingQuestion],
        strategy: Union[
            str, LabelQuestionStrategy, MultiLabelQuestionStrategy, RatingQuestionStrategy, RankingQuestionStrategy
        ],
    ) -> "FeedbackDataset":
        warnings.warn(
            "`unify_responses` method is deprecated and will be removed in future releases. "
            "Please use `compute_unified_responses` instead.",
            DeprecationWarning,
        )
        return self.compute_unified_responses(question=question, strategy=strategy)

    def compute_unified_responses(
        self: "FeedbackDatasetBase",
        question: Union[str, LabelQuestion, MultiLabelQuestion, RatingQuestion],
        strategy: Union[
            str, LabelQuestionStrategy, MultiLabelQuestionStrategy, RatingQuestionStrategy, RankingQuestionStrategy
        ],
    ) -> "FeedbackDataset":
        """
        The `compute_unified_responses` function takes a question and a strategy as input and applies the strategy
        to unify the responses for that question.

        Args:
            question The `question` parameter can be either a string representing the name of the
                question, or an instance of one of the question classes (`LabelQuestion`, `MultiLabelQuestion`,
                `RatingQuestion`, `RankingQuestion`).
            strategy The `strategy` parameter is used to specify the strategy to be used for unifying
                responses for a given question. It can be either a string or an instance of a strategy class.
        """
        if isinstance(question, str):
            question = self.question_by_name(question)

        if not strategy:
            strategy = "majority"

        if isinstance(strategy, str):
            if isinstance(question, LabelQuestion):
                strategy = LabelQuestionStrategy(strategy)
            elif isinstance(question, MultiLabelQuestion):
                strategy = MultiLabelQuestionStrategy(strategy)
            elif isinstance(question, RatingQuestion):
                strategy = RatingQuestionStrategy(strategy)
            elif isinstance(question, RankingQuestion):
                strategy = RankingQuestionStrategy(strategy)
            elif isinstance(question, TextQuestion):
                strategy = TextQuestionStrategy(strategy)
            else:
                raise ValueError(f"Question {question} is not supported yet")

        strategy.compute_unified_responses(self.records, question)
        return self



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TYPE_CHECKING

from argilla.client.feedback.training.frameworks.transformers import ArgillaTransformersTrainer
from argilla.training.peft import ArgillaPeftTrainer as ArgillaPeftTrainerV1
from argilla.utils.dependency import requires_dependencies

if TYPE_CHECKING:
    from argilla.client.feedback.integrations.huggingface.model_card import PeftModelCardData


class ArgillaPeftTrainer(ArgillaPeftTrainerV1, ArgillaTransformersTrainer):
    def __init__(self, *args, **kwargs):
        ArgillaTransformersTrainer.__init__(self, *args, **kwargs)

    def get_model_card_data(self, **card_data_kwargs) -> "PeftModelCardData":
        """
        Generate the card data to be used for the `ArgillaModelCard`.

        Args:
            card_data_kwargs: Extra arguments provided by the user when creating the `ArgillaTrainer`.

        Returns:
            PeftModelCardData: Container for the data to be written on the `ArgillaModelCard`.
        """
        from argilla.client.feedback.integrations.huggingface.model_card import PeftModelCardData

        return PeftModelCardData(
            model_id=self._model,
            task=self._task,
            update_config_kwargs=self.lora_kwargs,
            **card_data_kwargs,
        )

    @requires_dependencies("huggingface_hub")
    def push_to_huggingface(self, repo_id: str, **kwargs) -> None:
        """Uploads the model to [huggingface's model hub](https://huggingface.co/models).

        The full list of parameters can be seen at:
        [huggingface_hub](https://huggingface.co/docs/huggingface_hub/package_reference/mixins#huggingface_hub.ModelHubMixin.push_to_hub).

        Args:
            repo_id:
                The name of the repository you want to push your model and tokenizer to.
                It should contain your organization name when pushing to a given organization.
        """
        if not self.trainer_model:
            raise ValueError(
                "The model must be initialized prior to this point. You can either call `train` or `init_model`."
            )
        model_url = self.trainer_model.push_to_hub(repo_id, **kwargs)
        self._logger.info(f"Model pushed to: {model_url}")
        tokenizer_url = self.trainer_tokenizer.push_to_hub(repo_id, **kwargs)
        self._logger.info(f"Tokenizer pushed to: {tokenizer_url}")



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from typing import List, Union

from datasets import DatasetDict

from argilla.client.models import TextClassificationRecord, TokenClassificationRecord
from argilla.training.base import ArgillaTrainerSkeleton
from argilla.training.utils import _apply_column_mapping, filter_allowed_args, get_default_args
from argilla.utils.dependency import require_dependencies


class ArgillaTransformersTrainer(ArgillaTrainerSkeleton):
    _logger = logging.getLogger("ArgillaTransformersTrainer")
    _logger.setLevel(logging.INFO)

    def __init__(self, *args, **kwargs):
        require_dependencies(["torch", "datasets", "transformers", "evaluate", "seqeval"])
        super().__init__(*args, **kwargs)

        import torch
        from transformers import (
            AutoModelForSequenceClassification,
            AutoModelForTokenClassification,
            set_seed,
        )

        self.trainer_model = None
        self.trainer_tokenizer = None
        self.trainer_pipeline = None

        self.device = "cpu"
        if torch.backends.mps.is_available():
            self.device = "mps"
        elif torch.cuda.is_available():
            self.device = "cuda"

        if self._seed is None:
            self._seed = 42
        set_seed(self._seed)

        if self._model is None:
            self._model = "bert-base-cased"
            self._logger.warning(f"No model defined. Using the default model {self._model}.")

        if isinstance(self._dataset, DatasetDict):
            self._train_dataset = self._dataset["train"]
            self._eval_dataset = self._dataset["test"]
        else:
            self._train_dataset = self._dataset
            self._eval_dataset = None

        if self._record_class == TextClassificationRecord:
            if self._multi_label:
                self._id2label = dict(enumerate(self._train_dataset.features["label"][0].names))
                self._label_list = self._train_dataset.features["label"][0].names
            else:
                self._id2label = dict(enumerate(self._train_dataset.features["label"].names))
                self._label_list = self._train_dataset.features["label"].names
            self._label2id = {v: k for k, v in self._id2label.items()}

            self._model_class = AutoModelForSequenceClassification

        elif self._record_class == TokenClassificationRecord:
            self._label_list = self._train_dataset.features["ner_tags"].feature.names
            self._id2label = dict(enumerate(self._label_list))
            self._label2id = {v: k for k, v in self._id2label.items()}

            self._model_class = AutoModelForTokenClassification
        else:
            raise NotImplementedError("Text2TextRecord is not supported.")

        self.init_training_args()

    def init_training_args(self):
        import torch
        from transformers import TrainingArguments

        def convert_to_floats(dataset):
            dataset.set_format("torch")
            return dataset.map(
                lambda x: {"float_label": x["label"].to(torch.float)}, remove_columns=["label"]
            ).rename_column("float_label", "label")

        self.model_kwargs = {}
        self.model_kwargs["pretrained_model_name_or_path"] = (
            self._model if isinstance(self._model, str) else self._model.config._name_or_path
        )

        if self._record_class in [TextClassificationRecord, TokenClassificationRecord]:
            self.model_kwargs["num_labels"] = len(self._label_list)
            self.model_kwargs["id2label"] = self._id2label
            self.model_kwargs["label2id"] = self._label2id
            if self._multi_label:
                self.model_kwargs["problem_type"] = "multi_label_classification"

            if self._record_class == TextClassificationRecord:
                columns_mapping = {"text": "text", "label": "binarized_label"}
                if self._multi_label:
                    self._train_dataset = _apply_column_mapping(self._train_dataset, columns_mapping)
                    self._train_dataset = convert_to_floats(self._train_dataset)

                    if self._eval_dataset is not None:
                        self._eval_dataset = _apply_column_mapping(self._eval_dataset, columns_mapping)
                        self._eval_dataset = convert_to_floats(self._eval_dataset)

        self.trainer_kwargs = get_default_args(TrainingArguments.__init__)
        self.trainer_kwargs["evaluation_strategy"] = "no" if self._eval_dataset is None else "epoch"
        self.trainer_kwargs["logging_steps"] = 30
        self.trainer_kwargs["logging_steps"] = 1
        self.trainer_kwargs["num_train_epochs"] = 1

    def init_model(self, new: bool = False):
        from transformers import AutoTokenizer

        if self.trainer_tokenizer is None:
            if any(k in self.model_kwargs.get("pretrained_model_name_or_path") for k in ("gpt", "opt", "bloom")):
                padding_side = "left"
            else:
                padding_side = "right"

            self.trainer_tokenizer = AutoTokenizer.from_pretrained(
                self.model_kwargs.get("pretrained_model_name_or_path"),
                padding_side=padding_side,
                add_prefix_space=True,
            )
            if getattr(self.trainer_tokenizer, "pad_token_id") is None:
                self.trainer_tokenizer.pad_token_id = self.trainer_tokenizer.eos_token_id
            if getattr(self.trainer_tokenizer, "model_max_length") is None:
                self.trainer_tokenizer.model_max_length = 512

        if new:
            model_kwargs = self.model_kwargs
        else:
            model_kwargs = {"pretrained_model_name_or_path": self.model_kwargs.get("pretrained_model_name_or_path")}

        if self.trainer_model is None:
            self.trainer_model = self._model_class.from_pretrained(**model_kwargs, return_dict=True)
        if new:
            self.trainer_model = self.trainer_model.to(self.device)

    def init_pipeline(self):
        import transformers
        from transformers import AutoModelForQuestionAnswering, pipeline

        if self.device == "cuda":
            device = 0
        else:
            device = -1

        if self._record_class == TextClassificationRecord:
            if transformers.__version__ >= "4.20.0":
                kwargs = {"top_k": None}
            else:
                kwargs = {"return_all_scores": True}
            self.trainer_pipeline = pipeline(
                task="text-classification",
                model=self.trainer_model,
                tokenizer=self.trainer_tokenizer,
                device=device,
                **kwargs,
            )
        elif self._record_class == TokenClassificationRecord:
            self.trainer_pipeline = pipeline(
                task="token-classification",
                model=self.trainer_model,
                tokenizer=self.trainer_tokenizer,
                aggregation_strategy="first",
                device=device,
            )
        elif self._model_class == AutoModelForQuestionAnswering:
            self.trainer_pipeline = pipeline(
                task="question-answering",
                model=self.trainer_model,
                tokenizer=self.trainer_tokenizer,
                device=device,
            )
        else:
            raise NotImplementedError(f"{self._record_class} is not supported yet.")

    def update_config(self, **kwargs):
        """
        Updates the `setfit_model_kwargs` and `setfit_trainer_kwargs` dictionaries with the keyword
        arguments passed to the `update_config` function.
        """
        from transformers import TrainingArguments

        self.trainer_kwargs.update(filter_allowed_args(TrainingArguments.__init__, **kwargs))

    def __repr__(self):
        formatted_string = []
        arg_dict = {
            "'AutoModel'": self.model_kwargs,
            "'Trainer'": self.trainer_kwargs,
        }
        for arg_dict_key, arg_dict_single in arg_dict.items():
            formatted_string.append(arg_dict_key)
            for key, val in arg_dict_single.items():
                formatted_string.append(f"{key}: {val}")
        return "\n".join(formatted_string)

    def preprocess_datasets(self):
        from transformers import (
            AutoModelForQuestionAnswering,
            DataCollatorForTokenClassification,
            DataCollatorWithPadding,
            DefaultDataCollator,
        )

        def text_classification_preprocess_function(examples):
            return self.trainer_tokenizer(examples["text"], truncation=True, max_length=None)

        def token_classification_preprocess_function(examples):
            tokenized_inputs = self.trainer_tokenizer(examples["tokens"], truncation=True, is_split_into_words=True)

            labels = []
            for i, label in enumerate(examples["ner_tags"]):
                word_ids = tokenized_inputs.word_ids(batch_index=i)
                previous_word_idx = None
                label_ids = []
                for word_idx in word_ids:
                    if word_idx is None:
                        label_ids.append(-100)
                    elif word_idx != previous_word_idx:
                        label_ids.append(label[word_idx])
                    else:
                        label_ids.append(-100)
                    previous_word_idx = word_idx
                labels.append(label_ids)

            tokenized_inputs["labels"] = labels
            return tokenized_inputs

        def question_answering_preprocess_function(examples):
            questions = [q.strip() for q in examples["question"]]
            inputs = self.trainer_tokenizer(
                questions,
                examples["context"],
                truncation="only_second",
                return_offsets_mapping=True,
                padding="max_length",
            )

            offset_mapping = inputs.pop("offset_mapping")
            answers = examples["answer"]
            start_positions = []
            end_positions = []

            for i, offset in enumerate(offset_mapping):
                answer = answers[i]
                start_char = answer["answer_start"][0]
                end_char = answer["answer_start"][0] + len(answer["text"][0])
                sequence_ids = inputs.sequence_ids(i)

                # Find the start and end of the context
                idx = 0
                while sequence_ids[idx] != 1:
                    idx += 1
                context_start = idx
                while sequence_ids[idx] == 1:
                    idx += 1
                context_end = idx - 1

                # If the answer is not fully inside the context, label it (0, 0)
                if offset[context_start][0] > end_char or offset[context_end][1] < start_char:
                    start_positions.append(0)
                    end_positions.append(0)
                else:
                    # Otherwise it's the start and end token positions
                    idx = context_start
                    while idx <= context_end and offset[idx][0] <= start_char:
                        idx += 1
                    start_positions.append(idx - 1)

                    idx = context_end
                    while idx >= context_start and offset[idx][1] >= end_char:
                        idx -= 1
                    end_positions.append(idx + 1)

            inputs["start_positions"] = start_positions
            inputs["end_positions"] = end_positions
            return inputs

        def question_answering_preprocess_function_validation(examples):
            questions = [q.strip() for q in examples["question"]]
            inputs = self.trainer_tokenizer(
                questions,
                examples["context"],
                truncation="only_second",
                return_overflowing_tokens=True,
                return_offsets_mapping=True,
                padding="max_length",
            )

            sample_map = inputs.pop("overflow_to_sample_mapping")
            example_ids = []

            for i in range(len(inputs["input_ids"])):
                sample_idx = sample_map[i]
                example_ids.append(examples["id"][sample_idx])

                sequence_ids = inputs.sequence_ids(i)
                offset = inputs["offset_mapping"][i]
                inputs["offset_mapping"][i] = [o if sequence_ids[k] == 1 else None for k, o in enumerate(offset)]

            inputs["example_id"] = example_ids
            return inputs

        # set correct tokenization
        if self._record_class == TextClassificationRecord:
            preprocess_function = text_classification_preprocess_function
            self._data_collator = DataCollatorWithPadding(tokenizer=self.trainer_tokenizer)
            if self._multi_label:
                remove_columns = ["feat_id", "text", "feat_label"]
            else:
                remove_columns = ["id", "text"]
            replace_labels = True
        elif self._record_class == TokenClassificationRecord:
            preprocess_function = token_classification_preprocess_function
            self._data_collator = DataCollatorForTokenClassification(tokenizer=self.trainer_tokenizer)
            remove_columns = ["id", "tokens", "ner_tags"]
            replace_labels = False
        elif self._model_class == AutoModelForQuestionAnswering:
            preprocess_function = question_answering_preprocess_function
            self._data_collator = DefaultDataCollator()
            remove_columns = self._train_dataset.column_names
            replace_labels = False
        else:
            raise NotImplementedError("`Text2TextRecord` is not supported yet.")

        self._tokenized_train_dataset = self._train_dataset.map(
            preprocess_function, batched=True, remove_columns=remove_columns
        )
        if replace_labels:
            self._tokenized_train_dataset = self._tokenized_train_dataset.rename_column("label", "labels")

        if self._eval_dataset is not None:
            if self._model_class == AutoModelForQuestionAnswering:
                # We need to preprocess the validation dataset separately, because we need to return the example_id
                self._tokenized_eval_dataset = self._eval_dataset.map(
                    question_answering_preprocess_function_validation,
                    batched=True,
                    remove_columns=remove_columns,
                )
            else:
                self._tokenized_eval_dataset = self._eval_dataset.map(
                    preprocess_function, batched=True, remove_columns=remove_columns
                )

            if replace_labels:
                self._tokenized_eval_dataset = self._tokenized_eval_dataset.rename_column("label", "labels")
        else:
            self._tokenized_eval_dataset = None

    def compute_metrics(self):
        import collections

        import evaluate
        import numpy as np
        from transformers import AutoModelForQuestionAnswering

        func = None
        if self._record_class == TextClassificationRecord:
            accuracy = evaluate.load("accuracy")
            f1 = evaluate.load("f1", config_name="multilabel")

            def compute_metrics_text_classification_multi_label(eval_pred):
                logits, labels = eval_pred
                # apply sigmoid
                predictions = (1.0 / (1 + np.exp(-logits))) > 0.5

                # f1 micro averaged
                metrics = f1.compute(predictions=predictions, references=labels, average="micro")
                # f1 per label
                per_label_metric = f1.compute(predictions=predictions, references=labels, average=None)
                for label, f1_score in zip(self._label_list, per_label_metric["f1"]):
                    metrics[f"f1_{label}"] = f1_score

                return metrics

            def compute_metrics_text_classification(eval_pred):
                predictions, labels = eval_pred
                predictions = np.argmax(predictions, axis=1)
                return accuracy.compute(predictions=predictions, references=labels)

            if self._multi_label:
                func = compute_metrics_text_classification_multi_label
            else:
                func = compute_metrics_text_classification

        elif self._record_class == TokenClassificationRecord:
            seqeval = evaluate.load("seqeval")

            def compute_metrics(p):
                predictions, labels = p
                predictions = np.argmax(predictions, axis=2)

                true_predictions = [
                    [self._label_list[p] for (p, l) in zip(prediction, label) if l != -100]
                    for prediction, label in zip(predictions, labels)
                ]
                true_labels = [
                    [self._label_list[l] for (p, l) in zip(prediction, label) if l != -100]
                    for prediction, label in zip(predictions, labels)
                ]

                results = seqeval.compute(predictions=true_predictions, references=true_labels)
                return {
                    "precision": results["overall_precision"],
                    "recall": results["overall_recall"],
                    "f1": results["overall_f1"],
                    "accuracy": results["overall_accuracy"],
                }

            func = compute_metrics
        elif AutoModelForQuestionAnswering:
            squad = evaluate.load("squad")

            # Copy from https://huggingface.co/learn/nlp-course/chapter7/7?fw=pt#fine-tuning-the-model-with-the-trainer-api
            n_best = 20

            def compute_metrics_question_answering(pred):
                start_logits, end_logits = pred.predictions
                features = self._tokenized_eval_dataset
                examples = self._eval_dataset

                example_to_features = collections.defaultdict(list)
                for idx, feature in enumerate(features):
                    example_to_features[feature["example_id"]].append(idx)

                predicted_answers = []
                for example in examples:
                    example_id = example["id"]
                    context = example["context"]
                    answers = []

                    # Loop through all features associated with that example
                    for feature_index in example_to_features[example_id]:
                        start_logit = start_logits[feature_index]
                        end_logit = end_logits[feature_index]
                        offsets = features[feature_index]["offset_mapping"]

                        start_indexes = np.argsort(start_logit)[-1 : -n_best - 1 : -1].tolist()
                        end_indexes = np.argsort(end_logit)[-1 : -n_best - 1 : -1].tolist()
                        for start_index in start_indexes:
                            for end_index in end_indexes:
                                # Skip answers that are not fully in the context
                                if offsets[start_index] is None or offsets[end_index] is None:
                                    continue
                                # Skip answers with a length that is either < 0 or > max_answer_length
                                if (
                                    end_index < start_index
                                    or end_index - start_index + 1 > self.trainer_tokenizer.model_max_length
                                ):
                                    continue

                                answer = {
                                    "text": context[offsets[start_index][0] : offsets[end_index][1]],
                                    "logit_score": start_logit[start_index] + end_logit[end_index],
                                }
                                answers.append(answer)

                    # Select the answer with the best score
                    if len(answers) > 0:
                        best_answer = max(answers, key=lambda x: x["logit_score"])
                        predicted_answers.append({"id": example_id, "prediction_text": best_answer["text"]})
                    else:
                        predicted_answers.append({"id": example_id, "prediction_text": ""})

                theoretical_answers = [{"id": ex["id"], "answers": ex["answers"]} for ex in examples]

                return squad.compute(predictions=predicted_answers, references=theoretical_answers)

            func = compute_metrics_question_answering
        else:
            raise NotImplementedError("Other record types are not supported yet.")
        return func

    def train(self, output_dir: str):
        from transformers import Trainer, TrainingArguments

        # check required path argument
        self.trainer_kwargs["output_dir"] = output_dir

        # prepare data
        self.init_model(new=True)
        self.preprocess_datasets()

        # get metrics function
        compute_metrics = self.compute_metrics()

        self.trainer_model.to(self.device)
        # set trainer
        if self.device == "cuda":
            self.trainer_kwargs["no_cuda"] = False
        elif self.device == "mps":
            self.trainer_kwargs["use_mps_device"] = True
        self._trainer = Trainer(
            args=TrainingArguments(**self.trainer_kwargs),
            model=self.trainer_model,
            tokenizer=self.trainer_tokenizer,
            train_dataset=self._tokenized_train_dataset,
            eval_dataset=self._tokenized_eval_dataset,
            compute_metrics=compute_metrics,
            data_collator=self._data_collator,
        )

        #  train
        self._trainer.train()
        if self._tokenized_eval_dataset:
            self._metrics = self._trainer.evaluate()
            self._logger.info(self._metrics)
        else:
            self._metrics = None

        self.save(output_dir)

        self.init_pipeline()

    def predict(self, text: Union[List[str], str], as_argilla_records: bool = True, **kwargs):
        """
        The function takes in a list of strings and returns a list of predictions

        Args:
          text (Union[List[str], str]): The text to be classified.
          as_argilla_records (bool): If True, the prediction will be returned as an Argilla record. If
        False, the prediction will be returned as a string. Defaults to True

        Returns:
          A list of predictions
        """
        if self.trainer_pipeline is None:
            self._logger.warning("Using model without fine-tuning.")
            self.init_model(new=False)
            self.init_pipeline()

        str_input = False
        if isinstance(text, str):
            text = [text]
            str_input = True

        predictions = self.trainer_pipeline(text, **kwargs)

        if as_argilla_records:
            formatted_prediction = []

            for val, pred in zip(text, predictions):
                if self._record_class == TextClassificationRecord:
                    formatted_prediction.append(
                        self._record_class(
                            text=val,
                            prediction=[(entry["label"], entry["score"]) for entry in pred],
                            multi_label=self._multi_label,
                        )
                    )
                elif self._record_class == TokenClassificationRecord:
                    _pred = [(value["entity_group"], value["start"], value["end"]) for value in pred]
                    encoding = self.trainer_pipeline.tokenizer(val)
                    word_ids = sorted(set(encoding.word_ids()) - {None})
                    tokens = []
                    for word_id in word_ids:
                        char_span = encoding.word_to_chars(word_id)
                        tokens.append(val[char_span.start : char_span.end].lstrip().rstrip())
                    formatted_prediction.append(
                        self._record_class(
                            text=val,
                            tokens=tokens,
                            prediction=_pred,
                        )
                    )

                else:
                    raise NotImplementedError("This is not implemented yet.")
        else:
            formatted_prediction = predictions

        if str_input:
            formatted_prediction = formatted_prediction[0]

        return formatted_prediction

    def save(self, output_dir: str):
        """
        The function saves the model to the path specified and also saves the label2id and id2label
        dictionaries to the same path

        Args:
          output_dir (str): the path to save the model to
        """
        self.trainer_model.save_pretrained(output_dir)
        self.trainer_tokenizer.save_pretrained(output_dir)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import typing
from typing import Any, Dict, List, Optional, Type, Union

import httpx

from argilla.client.feedback.constants import FIELD_TYPE_TO_PYTHON_TYPE
from argilla.client.feedback.dataset.base import FeedbackDatasetBase
from argilla.client.feedback.schemas import FeedbackRecord
from argilla.client.feedback.schemas.enums import MetadataPropertyTypes
from argilla.client.feedback.schemas.types import AllowedFieldTypes, AllowedMetadataPropertyTypes, AllowedQuestionTypes
from argilla.client.sdk.v1.datasets import api as datasets_api_v1
from argilla.client.sdk.v1.datasets.models import FeedbackDatasetModel
from argilla.client.singleton import active_client
from argilla.client.workspaces import Workspace
from argilla.pydantic_v1 import BaseModel, Extra, ValidationError, create_model

if typing.TYPE_CHECKING:
    from argilla.client.feedback.schemas.types import (
        AllowedFieldTypes,
        AllowedMetadataPropertyTypes,
        AllowedRemoteFieldTypes,
        AllowedRemoteMetadataPropertyTypes,
    )


def validate_fields(fields: typing.Union[AllowedFieldTypes, typing.List[AllowedFieldTypes]]):
    """Validates that the fields used in the filters are valid."""
    if not isinstance(fields, list):
        raise TypeError(f"Expected `fields` to be a list, got {type(fields)} instead.")

    any_required = False
    unique_names = set()
    for field in fields:
        if not isinstance(field, AllowedFieldTypes):
            raise TypeError(
                f"Expected `fields` to be a list of `{AllowedFieldTypes.__name__}`, got {type(field)} instead."
            )
        if field.name in unique_names:
            raise ValueError(f"Expected `fields` to have unique names, got {field.name} twice instead.")
        unique_names.add(field.name)
        if not any_required and field.required:
            any_required = True

    if not any_required:
        raise ValueError("At least one field in `fields` must be required (`required=True`).")


def validate_questions(questions: typing.Union[AllowedQuestionTypes, typing.List[AllowedQuestionTypes]]) -> None:
    """Validates that the questions used in the filters are valid."""
    if not isinstance(questions, list):
        raise TypeError(f"Expected `questions` to be a list, got {type(questions)} instead.")

    any_required = False
    unique_names = set()
    for question in questions:
        if not isinstance(question, AllowedQuestionTypes.__args__):
            raise TypeError(
                "Expected `questions` to be a list of"
                f" `{'`, `'.join([arg.__name__ for arg in AllowedQuestionTypes.__args__])}` got a"
                f" question in the list with type {type(question)} instead."
            )
        if question.name in unique_names:
            raise ValueError(f"Expected `questions` to have unique names, got {question.name} twice instead.")
        unique_names.add(question.name)
        if not any_required and question.required:
            any_required = True

    if not any_required:
        raise ValueError("At least one question in `questions` must be required (`required=True`).")


def validate_metadata_properties(
    metadata_properties: typing.Union[typing.List[AllowedMetadataPropertyTypes], None]
) -> None:
    """Validates that the metadata properties used in the filters are valid."""

    if not metadata_properties:
        return

    unique_names = set()
    for metadata_property in metadata_properties:
        if not isinstance(metadata_property, AllowedMetadataPropertyTypes.__args__):
            raise TypeError(
                f"Expected `metadata_properties` to be a list of"
                f" `{'`, `'.join([arg.__name__ for arg in AllowedMetadataPropertyTypes.__args__])}` got a"
                f" metadata property in the list with type type {type(metadata_property)} instead."
            )
        if metadata_property.name in unique_names:
            raise ValueError(
                f"Expected `metadata_properties` to have unique names, got '{metadata_property.name}' twice instead."
            )
        unique_names.add(metadata_property.name)


def validate_metadata_names(dataset: "FeedbackDatasetBase", names: typing.List[str]) -> None:
    """Validates that the metadata names used in the filters are valid."""

    metadata_property_names = {metadata_property.name: True for metadata_property in dataset.metadata_properties}

    if not metadata_property_names:
        return

    for name in set(names):
        if not metadata_property_names.get(name):
            raise ValueError(
                f"The metadata property name `{name}` does not exist in the current `FeedbackDataset` in Argilla."
                f" The existing metadata properties names are: {list(metadata_property_names.keys())}."
            )


def validate_vector_names(dataset: "FeedbackDatasetBase", names: typing.List[str]) -> None:
    """Validates that the vector names used in the filters are valid."""
    vectors_names = {vector.name: True for vector in dataset.vectors_settings}

    if not vectors_names:
        return

    for name in set(names):
        if not vectors_names.get(name):
            raise ValueError(
                f"The vector name `{name}` does not exist in the current `FeedbackDataset` in Argilla."
                f" The existing vector names are: {list(vectors_names.keys())}."
            )


def normalize_records(
    records: Union[FeedbackRecord, Dict[str, Any], List[Union[FeedbackRecord, Dict[str, Any]]]]
) -> List[FeedbackRecord]:
    """Parses the records into a list of `FeedbackRecord` objects.

    Args:
        records: either a single `FeedbackRecord` or `dict` or a list of `FeedbackRecord` or `dict`.

    Returns:
        A list of `FeedbackRecord` objects.

    Raises:
        ValueError: if `records` is not a `FeedbackRecord` or `dict` or a list of `FeedbackRecord` or `dict`.
    """
    if isinstance(records, (dict, FeedbackRecord)):
        records = [records]

    if len(records) == 0:
        raise ValueError("Expected `records` to be a non-empty list of `dict` or `FeedbackRecord`.")

    new_records = []
    for record in records:
        if isinstance(record, dict):
            new_records.append(FeedbackRecord(**record))
        elif isinstance(record, FeedbackRecord):
            new_records.append(record)
        else:
            raise ValueError(
                "Expected `records` to be a list of `dict` or `FeedbackRecord`," f" got type `{type(record)}` instead."
            )
    return new_records


def validate_dataset_records(
    dataset: FeedbackDatasetBase,
    records: List[FeedbackRecord],
    attributes_to_validate: typing.Optional[List[str]] = None,
) -> None:
    """Validates the records against the schema defined by the `fields`.

    Args:
        dataset: the `FeedbackDataset` object to validate the records against.
        records: a list of `FeedbackRecord` objects to validate.
        attributes_to_validate: a list containing the name of the attributes to
            validate from the record. Valid values are: `fields`, `metadata` and `vectors`.
            If not provided, all `fields`, `metadata` and `vectors` are validated. Defaults
            to `None`.

    Raises:
        ValueError: if the `fields` schema does not match the `FeedbackRecord.fields` schema.
    """
    attributes_to_validate = attributes_to_validate or ["fields", "metadata", "vectors"]

    if "fields" in attributes_to_validate:
        fields_schema = _build_fields_schema(dataset)

    if "metadata" in attributes_to_validate:
        metadata_schema = _build_metadata_schema(dataset)

    if "vectors" in attributes_to_validate:
        vectors_settings_by_name = {
            vector_settings.name: vector_settings for vector_settings in dataset.vectors_settings or []
        }

    for record in records:
        if "fields" in attributes_to_validate:
            _validate_record_fields(record, fields_schema)

        if "metadata" in attributes_to_validate:
            _validate_record_metadata(record, metadata_schema)

        if "vectors" in attributes_to_validate:
            _validate_record_vectors(record, vectors_settings_by_name)


def get_dataset_by_name_and_workspace(
    name: Optional[str] = None, *, workspace: Optional[Union[str, Workspace]] = None, id: Optional[str] = None
) -> Optional["FeedbackDatasetModel"]:
    """Checks whether a `FeedbackDataset` exists in Argilla or not, based on the `name`, `id`, or the combination of
    `name` and `workspace`.

    Args:
        name: the name of the `FeedbackDataset` in Argilla.
        workspace: the name of the workspace in Argilla where the `FeedbackDataset` is located.
        id: the Argilla ID of the `FeedbackDataset`.

    Returns:
        The `FeedbackDataset` if it exists in Argilla, `None` otherwise.

    Raises:
        ValueError: if the `workspace` is not a `Workspace` instance or a string.
        Exception: if the `FeedbackDataset` could not be listed from Argilla.

    Examples:
        >>> import argilla as rg
        >>> rg.init(api_url="...", api_key="...")
        >>> from argilla.client.feedback.dataset.helpers import get_dataset_by_name_and_workspace
        >>> dataset = get_dataset_by_name_and_workspace(name="my-dataset")
    """
    assert (name and workspace) or name or id, (
        "You must provide either the `name` and `workspace` (the latter just if"
        " applicable, if not the default `workspace` will be used) or the `id`, which"
        " is the Argilla ID of the `rg.FeedbackDataset`."
    )

    client = active_client()
    httpx_client: "httpx.Client" = client.http_client.httpx

    if name:
        if workspace is None:
            workspace = Workspace.from_name(client.get_workspace())
        elif isinstance(workspace, str):
            workspace = Workspace.from_name(workspace)
        elif not isinstance(workspace, Workspace):
            raise ValueError(f"Workspace must be a `rg.Workspace` instance or a string, got {type(workspace)}")

        try:
            datasets = datasets_api_v1.list_datasets(client=httpx_client).parsed
        except Exception as e:
            raise Exception(f"Failed while listing the `FeedbackTask` datasets from Argilla with exception: {e}")

        for dataset in datasets:
            if dataset.name == name and dataset.workspace_id == workspace.id:
                return dataset
        return None
    elif id:
        try:
            return datasets_api_v1.get_dataset(client=httpx_client, id=id).parsed
        except Exception:
            return None
    else:
        raise ValueError("You must provide either the `name` and `workspace` or the `id` of the `FeedbackDataset`.")


def generate_pydantic_schema_for_fields(
    fields: List[Union["AllowedFieldTypes", "AllowedRemoteFieldTypes"]], name: Optional[str] = "FieldsSchema"
) -> typing.Type[BaseModel]:
    """Generates a `pydantic.BaseModel` schema from a list of `AllowedFieldTypes` or `AllowedRemoteFieldTypes`
    objects to validate the fields of a `FeedbackDataset` or `RemoteFeedbackDataset` object, respectively,
    before inserting them.

    Args:
        fields: the list of `AllowedFieldTypes` or `AllowedRemoteFieldTypes` objects to generate the schema from.
        name: the name of the `pydantic.BaseModel` schema to generate. Defaults to "FieldsSchema".

    Returns:
        A `pydantic.BaseModel` schema to validate the fields of a `FeedbackDataset` or `RemoteFeedbackDataset`
        object before inserting them.

    Raises:
        ValueError: if one of the fields has an unsupported type.

    Examples:
        >>> from argilla.client.feedback.schemas.fields import TextField
        >>> from argilla.client.feedback.dataset.helpers import generate_pydantic_schema_for_fields
        >>> fields = [
        ...     TextField(name="text", required=True),
        ...     TextField(name="label", required=True),
        ... ]
        >>> FieldsSchema = generate_pydantic_schema_for_fields(fields)
        >>> FieldsSchema(text="Hello", label="World")
        FieldsSchema(text='Hello', label='World')
    """
    fields_schema = {}
    for field in fields:
        if field.type not in FIELD_TYPE_TO_PYTHON_TYPE.keys():
            raise ValueError(
                f"Field {field.name} has an unsupported type: {field.type}, for the moment only the"
                f" following types are supported: {list(FIELD_TYPE_TO_PYTHON_TYPE.keys())}"
            )
        fields_schema.update({field.name: (FIELD_TYPE_TO_PYTHON_TYPE[field.type], ... if field.required else None)})
    return create_model(name, **fields_schema)


def generate_pydantic_schema_for_metadata(
    metadata_properties: List[Union["AllowedMetadataPropertyTypes", "AllowedRemoteMetadataPropertyTypes"]],
    name: Optional[str] = "MetadataSchema",
    allow_extra_metadata: bool = True,
) -> Type[BaseModel]:
    """Generates a `pydantic.BaseModel` schema from a list of `AllowedMetadataPropertyTypes` or
    `AllowedRemoteMetadataPropertyTypes` objects to validate the metadata of a `FeedbackDataset`
    or `RemoteFeedbackDataset` object, respectively, before inserting them.

    Args:
        metadata_properties: the list of `AllowedMetadataPropertyTypes` or `AllowedRemoteMetadataPropertyTypes`
            objects to generate the schema from.
        name: the name of the `pydantic.BaseModel` schema to generate. Defaults to "MetadataSchema".
        allow_extra_metadata: whether to allow extra metadata properties or not. Defaults to `True`.

    Returns:
        A `pydantic.BaseModel` schema to validate the metadata of a `FeedbackDataset` or `RemoteFeedbackDataset`
        object before inserting them.

    Raises:
        ValueError: if one of the metadata properties has an unsupported type.

    Examples:
        >>> from argilla.client.feedback.schemas.metadata import IntegerMetadataProperty
        >>> from argilla.client.feedback.dataset.helpers import generate_pydantic_schema_for_metadata
        >>> metadata_properties = [
        ...     IntegerMetadataProperty(name="int_metadata", min=0, max=10),
        ...     ...,
        ... ]
        >>> MetadataSchema = generate_pydantic_schema_for_metadata(metadata_properties)
        >>> MetadataSchema(int_metadata=5)
        MetadataSchema(int-metadata=5)
    """
    metadata_fields, metadata_validators = {}, {}

    for metadata_property in metadata_properties or []:
        if metadata_property.type not in [arg.value for arg in MetadataPropertyTypes]:
            raise ValueError(
                f"Metadata property {metadata_property.name} has an unsupported type: {metadata_property.type}, for the moment only the"
                f" following types are supported: {[arg.value for arg in MetadataPropertyTypes]}"
            )
        pydantic_field, pydantic_validator = metadata_property._pydantic_field_with_validator
        metadata_fields.update(pydantic_field)
        metadata_validators.update(pydantic_validator)

    class MetadataConfig:
        extra = Extra.allow if allow_extra_metadata else Extra.forbid

    return create_model(name, **metadata_fields, __validators__=metadata_validators, __config__=MetadataConfig)


def _validate_record_fields(record: FeedbackRecord, fields_schema: typing.Type[BaseModel]) -> None:
    """Validates the `FeedbackRecord.fields` against the schema defined by the `fields`."""
    try:
        fields_schema.parse_obj(record.fields)
    except ValidationError as e:
        raise ValueError(f"`FeedbackRecord.fields` does not match the expected schema, with exception: {e}") from e


def _validate_record_metadata(record: FeedbackRecord, metadata_schema: typing.Type[BaseModel] = None) -> None:
    """Validates the `FeedbackRecord.metadata` against the schema defined by the `metadata_properties`."""

    if not record.metadata:
        return

    try:
        metadata_schema.parse_obj(record.metadata)
    except ValidationError as e:
        raise ValueError(
            f"`FeedbackRecord.metadata` {record.metadata} does not match the expected schema," f" with exception: {e}"
        ) from e


def _validate_record_vectors(record: FeedbackRecord, vectors_settings_by_name) -> None:
    for vector_name, vector_value in record.vectors.items():
        if not vectors_settings_by_name.get(vector_name):
            raise ValueError(f"Vector with name `{vector_name}` not present on dataset vector settings.")

        if vectors_settings_by_name[vector_name].dimensions != (record_vector_dimension := len(vector_value)):
            raise ValueError(
                f"Vector with name `{vector_name}` has an invalid expected dimension. {record_vector_dimension}"
            )

        if not (isinstance(vector_value, list) and all(isinstance(v, float) for v in vector_value)):
            raise ValueError(
                f"Vector with name `{vector_name}` has an invalid type. "
                f"Expected `list[float]`, got `{type(vector_value)}`."
            )


def _build_fields_schema(dataset: FeedbackDatasetBase) -> Type[BaseModel]:
    """Returns the fields schema of the dataset."""
    return generate_pydantic_schema_for_fields(dataset.fields)


def _build_metadata_schema(dataset: FeedbackDatasetBase) -> Type[BaseModel]:
    """Returns the metadata schema of the dataset."""
    return generate_pydantic_schema_for_metadata(
        dataset.metadata_properties, allow_extra_metadata=dataset.allow_extra_metadata
    )



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TYPE_CHECKING, List, Optional, Union

from argilla.client.sdk.v1.datasets import api as datasets_api_v1
from argilla.client.sdk.v1.datasets.models import FeedbackMetadataPropertyModel
from argilla.client.sdk.v1.metadata_properties import api as metadata_properties_api_v1

if TYPE_CHECKING:
    from uuid import UUID

    import httpx

    from argilla.client.feedback.schemas.types import AllowedMetadataPropertyTypes, AllowedRemoteMetadataPropertyTypes


class ArgillaMetadataPropertiesMixin:
    @staticmethod
    def parse_payload(
        client: "httpx.Client", payload: Union[FeedbackMetadataPropertyModel, List[FeedbackMetadataPropertyModel]]
    ) -> List["AllowedRemoteMetadataPropertyTypes"]:
        from argilla.client.feedback.schemas.remote.metadata import RemoteMetadataPropertiesMapping

        if isinstance(payload, FeedbackMetadataPropertyModel):
            payload = [payload]

        metadata_properties = list()
        for metadata_property in payload:
            metadata_properties.append(
                RemoteMetadataPropertiesMapping[metadata_property.settings["type"]].from_api(
                    payload=metadata_property,
                    client=client,
                )
            )
        return metadata_properties

    @staticmethod
    def list(client: "httpx.Client", dataset_id: "UUID") -> List["AllowedRemoteMetadataPropertyTypes"]:
        """Returns the list of metadata properties allowed for the given dataset, if any.

        Args:
            client: contains the `httpx.Client` instance that will be used to send requests to Argilla.
            dataset_id: contains the UUID of the dataset in Argilla.

        Returns:
            A list with the allowed metadata properties for the given dataset, if any.
        """
        response = datasets_api_v1.get_metadata_properties(client=client, id=dataset_id)
        if response.status_code == 200:
            return ArgillaMetadataPropertiesMixin.parse_payload(client=client, payload=response.parsed)

    @staticmethod
    def delete(client: "httpx.Client", metadata_property_id: "UUID") -> "AllowedMetadataPropertyTypes":
        """Deletes the metadata property with the given ID from Argilla.

        Args:
            client: contains the `httpx.Client` instance that will be used to send requests to Argilla.
            metadata_property_id: contains the UUID of the metadata property in Argilla.

        Returns:
            The deleted `RemoteMetadataProperty` as a `MetadataProperty` object.
        """
        response = metadata_properties_api_v1.delete_metadata_property(client=client, id=metadata_property_id)
        if response.status_code == 200:
            return ArgillaMetadataPropertiesMixin.parse_payload(client=client, payload=response.parsed)[0].to_local()

    @staticmethod
    def update(
        client: "httpx.Client",
        metadata_property_id: "UUID",
        title: Optional[str] = None,
        visible_for_annotators: Optional[bool] = None,
    ) -> "AllowedRemoteMetadataPropertyTypes":
        """Updates the metadata property with the given ID in Argilla.

        Args:
            client: contains the `httpx.Client` instance that will be used to send requests to Argilla.
            metadata_property_id: contains the UUID of the metadata property in Argilla.
            title: the new title of the metadata property. Defaults to `None`.
            visible_for_annotators: whether the metadata property should be visible for annotators. Defaults to `None`.

        Returns:
            The updated `RemoteMetadataProperty` object.
        """
        response = metadata_properties_api_v1.update_metadata_property(
            client=client,
            id=metadata_property_id,
            title=title,
            visible_for_annotators=visible_for_annotators,
        )
        if response.status_code == 200:
            return ArgillaMetadataPropertiesMixin.parse_payload(client=client, payload=response.parsed)[0]



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, Dict, Optional

from argilla.client.client import Argilla


class ArgillaSingleton:
    """The active argilla singleton instance"""

    _INSTANCE: Optional[Argilla] = None

    @classmethod
    def get(cls) -> Argilla:
        if cls._INSTANCE is None:
            return cls.init()
        return cls._INSTANCE

    @classmethod
    def clear(cls) -> None:
        cls._INSTANCE = None

    @classmethod
    def init(
        cls,
        api_url: Optional[str] = None,
        api_key: Optional[str] = None,
        workspace: Optional[str] = None,
        timeout: int = 60,
        extra_headers: Optional[Dict[str, str]] = None,
        httpx_extra_kwargs: Optional[Dict[str, Any]] = None,
    ) -> Argilla:
        cls._INSTANCE = None

        cls._INSTANCE = Argilla(
            api_url=api_url,
            api_key=api_key,
            timeout=timeout,
            workspace=workspace,
            extra_headers=extra_headers,
            httpx_extra_kwargs=httpx_extra_kwargs,
        )

        return cls._INSTANCE


def init(
    api_url: Optional[str] = None,
    api_key: Optional[str] = None,
    workspace: Optional[str] = None,
    timeout: int = 60,
    extra_headers: Optional[Dict[str, str]] = None,
    httpx_extra_kwargs: Optional[Dict[str, Any]] = None,
) -> None:
    """Init the Python client.

    If this function is called with `api_url=None` and `api_key=None` and no values have been set for the environment
    variables `ARGILLA_API_URL` and `ARGILLA_API_KEY`, then the local credentials stored by a previous call to `argilla
    login` command will be used. If local credentials are not found, then `api_url` and `api_key` will fallback to the
    default values.

    Args:
        api_url: Address of the REST API. If `None` (default) and the env variable ``ARGILLA_API_URL`` is not set,
            it will default to `http://localhost:6900`.
        api_key: Authentication key for the REST API. If `None` (default) and the env variable ``ARGILLA_API_KEY``
            is not set, it will default to `argilla.apikey`.
        workspace: The workspace to which records will be logged/loaded. If `None` (default) and the
            env variable ``ARGILLA_WORKSPACE`` is not set, it will default to the private user workspace.
        timeout: Wait `timeout` seconds for the connection to timeout. Default: 60.
        extra_headers: Extra HTTP headers sent to the server. You can use this to customize
            the headers of argilla client requests, like additional security restrictions. Default: `None`.
        httpx_extra_kwargs: Extra kwargs passed to the `httpx.Client` constructor. For more information about the
            available arguments, see https://www.python-httpx.org/api/#client. Defaults to `None`.

    Examples:
        >>> import argilla as rg
        >>>
        >>> rg.init(api_url="http://localhost:9090", api_key="4AkeAPIk3Y")
        >>> # Customizing request headers
        >>> headers = {"X-Client-id":"id","X-Secret":"secret"}
        >>> rg.init(api_url="http://localhost:9090", api_key="4AkeAPIk3Y", extra_headers=headers)
    """
    ArgillaSingleton.init(
        api_url=api_url,
        api_key=api_key,
        workspace=workspace,
        timeout=timeout,
        extra_headers=extra_headers,
        httpx_extra_kwargs=httpx_extra_kwargs,
    )


def active_client() -> Argilla:
    """Returns the active argilla client.

    If Active client is None, initialize a default one.
    """
    return ArgillaSingleton.get()


active_api = active_client  # backward compatibility



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import warnings
from typing import Optional, Union

from ..client.singleton import active_api
from ._flair import SequenceTagger, flair_monitor
from ._spacy import Language, ner_monitor
from ._transformers import Pipeline, huggingface_monitor
from .base import BaseMonitor

_LOGGER = logging.getLogger(__name__)


def monitor(
    task_model: Union[Language, Pipeline, SequenceTagger],
    dataset: str,
    sample_rate: float = 0.3,
    agent: Optional[str] = None,
    log_interval: float = 5,
) -> Union[BaseMonitor, Language, Pipeline, SequenceTagger]:
    """Automatically monitor (i.e. log) data fed through Transformer pipelines,
    spaCy models or flAIr taggers.

    Args:
        task_model (Union[Language, Pipeline, SequenceTagger]): The spaCy `Language`,
            transformers `Pipeline` or flAIr `SequenceTagger`.
        dataset (str): The Argilla dataset to log data into.
        sample_rate (float, optional): The portion of processed data to log. Defaults to 0.3.
        agent (Optional[str], optional): The name of the logging agent. Defaults to None.
        log_interval (float, optional): The interval for uploading in seconds. Defaults to 5.

    Returns:
        Union[BaseMonitor, Language, Pipeline, SequenceTagger]: The monitor that acts equivalently
            to the input task_model.
    """
    model_monitor = None
    api = active_api()
    if isinstance(task_model, Language):
        model_monitor = ner_monitor(
            task_model,
            api=api,
            dataset=dataset,
            sample_rate=sample_rate,
            log_interval=log_interval,
        )
    elif isinstance(task_model, Pipeline):
        model_monitor = huggingface_monitor(
            task_model,
            api=api,
            dataset=dataset,
            sample_rate=sample_rate,
            log_interval=log_interval,
        )
    elif isinstance(task_model, SequenceTagger):
        model_monitor = flair_monitor(
            task_model,
            api=api,
            dataset=dataset,
            sample_rate=sample_rate,
            log_interval=log_interval,
        )
    if model_monitor:
        model_monitor.agent = agent
        return model_monitor

    warnings.warn(
        "The provided task model is not supported by monitoring module. Predictions won't be logged into argilla."
    )
    return task_model



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from typing import List, Union

import numpy as np

from argilla.client.models import TextClassificationRecord, TokenClassificationRecord
from argilla.training.transformers import ArgillaTransformersTrainer
from argilla.training.utils import filter_allowed_args
from argilla.utils.dependency import require_dependencies


class ArgillaPeftTrainer(ArgillaTransformersTrainer):
    _logger = logging.getLogger("ArgillaTransformersPEFTTrainer")
    _logger.setLevel(logging.INFO)

    import sys

    if sys.version_info < (3, 9):
        raise Exception("Must be using Python 3.9 or higher or PEFT won't work.")

    def init_training_args(self):
        require_dependencies("peft")
        super().init_training_args()

        self.lora_kwargs = {
            "r": 8,
            "target_modules": None,
            "lora_alpha": 16,
            "lora_dropout": 0.1,
            "fan_in_fan_out": False,
            "bias": "none",
            "inference_mode": False,
            "modules_to_save": None,
            "init_lora_weights": True,
        }

    def init_model(self, new: bool = False):
        from peft import LoraConfig, PeftConfig, PeftModel, get_peft_model
        from transformers import AutoTokenizer

        if self._record_class == TextClassificationRecord:
            self.lora_kwargs["task_type"] = "SEQ_CLS"
        elif self._record_class == TokenClassificationRecord:
            self.lora_kwargs["task_type"] = "TOKEN_CLS"
        else:
            raise NotImplementedError("`rg.Text2TextRecord` is not supported yet.")

        try:
            config = PeftConfig.from_pretrained(self.model_kwargs["pretrained_model_name_or_path"])
            model = self._model_class.from_pretrained(
                config.base_model_name_or_path,
                return_dict=True,
                num_labels=len(self._id2label),
                id2label=self._id2label,
                label2id=self._label2id,
            )
            self._model_sub_class = model.__class__
            model = PeftModel.from_pretrained(model, self.model_kwargs["pretrained_model_name_or_path"])
        except Exception:
            config = LoraConfig(**self.lora_kwargs)
            model = self._model_class.from_pretrained(**self.model_kwargs, return_dict=True)
            self._model_sub_class = model.__class__
            model = get_peft_model(model, config)
        self.trainer_tokenizer = AutoTokenizer.from_pretrained(config.base_model_name_or_path, add_prefix_space=True)
        self.trainer_model = model.to(self.device)

    def init_pipeline(self):
        pass

    def update_config(self, **kwargs):
        """
        Updates the `model_kwargs` and `trainer_kwargs` dictionaries with the keyword
        arguments passed to the `update_config` function.
        """
        super().update_config(**kwargs)

        from peft import LoraConfig

        self.lora_kwargs.update(filter_allowed_args(LoraConfig.__init__, **kwargs))

    def __repr__(self):
        formatted_string = []
        arg_dict = {
            "'AutoModel'": self.model_kwargs,
            "'Trainer'": self.trainer_kwargs,
            "'LoraConfig'": self.lora_kwargs,
        }
        for arg_dict_key, arg_dict_single in arg_dict.items():
            formatted_string.append(arg_dict_key)
            for key, val in arg_dict_single.items():
                formatted_string.append(f"{key}: {val}")
        return "\n".join(formatted_string)

    def predict(self, text: Union[List[str], str], as_argilla_records: bool = True, **kwargs):
        """
        The function takes in a list of strings and returns a list of predictions

        Args:
          text (Union[List[str], str]): The text to be classified.
          as_argilla_records (bool): If True, the prediction will be returned as an Argilla record. If
        False, the prediction will be returned as a string. Defaults to True

        Returns:
          A list of predictions
        """
        import torch

        if self.trainer_model is None:
            self._logger.warning("Using model without fine-tuning.")
            self.init_model(new=False)

        str_input = False
        if isinstance(text, str):
            text = [text]
            str_input = True

        if self._record_class == TextClassificationRecord:
            inputs = self.trainer_tokenizer(text, truncation=True, padding="longest", return_tensors="pt")

            with torch.no_grad():
                logits = self.trainer_model(**inputs).logits

            if self._multi_label:
                probabilities = torch.sigmoid(logits)
            else:
                probabilities = torch.softmax(logits, dim=1)
            predictions = []
            for probability in probabilities:
                prediction = []
                for idx, score in enumerate(probability):
                    prediction.append({"label": self._id2label[idx], "score": score})
                predictions.append(prediction)
        else:
            # Tokenize the text
            inputs_with_offsets = self.trainer_tokenizer(
                text, truncation=True, padding="longest", return_offsets_mapping=True, return_tensors="pt"
            )
            inputs_with_offsets = {k: v.to(self.device) for k, v in inputs_with_offsets.items()}
            offsets = inputs_with_offsets["offset_mapping"]

            # Perform the forward pass through the model
            with torch.no_grad():
                outputs = self.trainer_model(**{k: v for k, v in inputs_with_offsets.items() if k != "offset_mapping"})
                probabilities = torch.nn.functional.softmax(outputs.logits, dim=-1).tolist()
                predictions = outputs.logits.argmax(dim=-1).tolist()

            # Iterate over the predictions and the offsets
            batch_formatted_predictions = []
            for batch_idx, example in enumerate(text):
                formatted_predictions = []
                idx = 0
                while idx < len(predictions[batch_idx]):
                    pred = predictions[batch_idx][idx]
                    label = self.trainer_model.config.id2label[pred]
                    if label != "O":
                        # Remove the B- or I-
                        label = label[2:]
                        start, end = offsets[batch_idx][idx]

                        # Grab all the tokens labeled with I-label
                        all_scores = []
                        while (
                            idx < len(predictions[batch_idx])
                            and self.trainer_model.config.id2label[predictions[batch_idx][idx]] == f"I-{label}"
                        ):
                            all_scores.append(probabilities[batch_idx][idx][pred])
                            _, end = offsets[batch_idx][idx]
                            idx += 1

                        # If the entity is only one token, we don't need to do anything
                        if start != end:
                            # The score is the mean of all the scores of the tokens in that grouped entity
                            score = None if np.isnan(np.mean(all_scores).item()) else np.mean(all_scores).item()
                            word = example[start:end]
                            formatted_predictions.append(
                                {
                                    "entity_group": label,
                                    "score": score,
                                    "word": word,
                                    "start": int(start),
                                    "end": int(end),
                                }
                            )
                    idx += 1
                batch_formatted_predictions.append(formatted_predictions)
            predictions = batch_formatted_predictions

        if as_argilla_records:
            formatted_prediction = []

            for val, pred in zip(text, predictions):
                if self._record_class == TextClassificationRecord:
                    formatted_prediction.append(
                        self._record_class(
                            text=val,
                            prediction=[(entry["label"], entry["score"]) for entry in pred],
                            multi_label=self._multi_label,
                        )
                    )
                elif self._record_class == TokenClassificationRecord:
                    formatted_prediction.append(
                        self._record_class(
                            text=val,
                            tokens=val.split(),
                            prediction=[
                                (
                                    char_spans["entity_group"],
                                    char_spans["start"],
                                    char_spans["end"],
                                    char_spans["score"],
                                )
                                for char_spans in pred
                            ],
                        )
                    )

                else:
                    raise NotImplementedError("`rg.Text2TextRecord` is not supported yet.")
        else:
            formatted_prediction = predictions

        if str_input:
            formatted_prediction = formatted_prediction[0]

        return formatted_prediction

    def save(self, output_dir: str):
        """
        The function saves the model to the path specified and also saves the label2id and id2label
        dictionaries to the same path

        Args:
          output_dir (str): the path to save the model to
        """
        self.trainer_model.save_pretrained(output_dir)
        self.trainer_tokenizer.save_pretrained(output_dir)



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
This module contains the data models for the interface
"""
import datetime
import logging
import uuid
import warnings
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple, Union

import pandas as pd
from deprecated import deprecated

from argilla import _messages
from argilla._constants import DEFAULT_MAX_KEYWORD_LENGTH
from argilla.pydantic_v1 import BaseModel, Field, PrivateAttr, root_validator, validator
from argilla.utils.span_utils import SpanUtils

_LOGGER = logging.getLogger(__name__)

Vectors = Dict[str, List[float]]


FRAMEWORK_TO_NAME_MAPPING = {
    "transformers": "Transformers",
    "peft": "PEFT Transformers library",
    "setfit": "SetFit Transformers library",
    "spacy": "Spacy Explosion",
    "spacy-transformers": "Spacy Transformers Explosion library",
    "span_marker": "SpanMarker Tom Aarsen library",
    "spark-nlp": "Spark NLP John Snow Labs library",
    "openai": "OpenAI LLMs",
    "trl": "Transformer Reinforcement Learning",
    "sentence-transformers": "Sentence Transformers library",
}


class Framework(Enum):
    """Frameworks supported by Argilla

    Options:
        transformers: Transformers
        peft: PEFT Transformers library
        setfit: SetFit Transformers library
        spacy: Spacy Explosion
        spacy-transformers: Spacy Transformers Explosion library
        span_marker: SpanMarker Tom Aarsen library
        spark-nlp: Spark NLP John Snow Labs library
        openai: OpenAI LLMs
        trl: Transformer Reinforcement Learning
        sentence-transformers: Sentence Transformers library
    """

    TRANSFORMERS = "transformers"
    PEFT = "peft"
    SETFIT = "setfit"
    SPACY = "spacy"
    SPACY_TRANSFORMERS = "spacy-transformers"
    SPAN_MARKER = "span_marker"
    SPARK_NLP = "spark-nlp"
    OPENAI = "openai"
    TRL = "trl"
    SENTENCE_TRANSFORMERS = "sentence-transformers"
    # AUTOTRAIN = "autotrain"

    @classmethod
    def _missing_(cls, value):
        raise ValueError(
            f"{value!r} is not a valid {cls.__name__.lower()}, please select one of {list(cls._value2member_map_.keys())}"
        )

    def __str__(self) -> str:
        return self.value


class _Validators(BaseModel):
    """Base class for our record models that takes care of general validations"""

    # The metadata field name prefix defined for protected (non-searchable) values
    _PROTECTED_METADATA_FIELD_PREFIX = "_"

    _JS_MAX_SAFE_INTEGER = 9007199254740991

    @validator("metadata", check_fields=False)
    def _check_value_length(cls, metadata):
        """Checks metadata values length and warn message for large values"""
        if not metadata:
            return metadata

        default_length_exceeded = False
        for k, v in metadata.items():
            if k.startswith(cls._PROTECTED_METADATA_FIELD_PREFIX):
                continue
            if isinstance(v, str) and len(v) > DEFAULT_MAX_KEYWORD_LENGTH:
                default_length_exceeded = True
                break

        if default_length_exceeded:
            message = (
                "Some metadata values could exceed the max length. For those cases,"
                " values will be truncated by keeping only the last"
                f" {DEFAULT_MAX_KEYWORD_LENGTH} characters. " + _messages.ARGILLA_METADATA_FIELD_WARNING_MESSAGE
            )
            warnings.warn(message, UserWarning, stacklevel=2)

        return metadata

    @validator("metadata", check_fields=False)
    def _none_to_empty_dict(cls, v):
        if v is None:
            return {}
        return v

    @validator("id", check_fields=False, pre=True, always=True)
    def _normalize_id(cls, v):
        if v is None:
            return str(uuid.uuid4())
        if isinstance(v, int):
            message = (
                "Integer ids won't be supported in future versions. We recommend to start using strings instead. "
                "For datasets already containing integer values we recommend migrating them to avoid deprecation issues. "
                "See https://docs.argilla.io/en/latest/getting_started/installation/configurations"
                "/database_migrations.html#elasticsearch"
            )
            warnings.warn(message, DeprecationWarning, stacklevel=2)
            # See https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Number/MAX_SAFE_INTEGER
            if v > cls._JS_MAX_SAFE_INTEGER:
                message = (
                    "You've provided a big integer value. Use a string instead, otherwise you may experience some "
                    "problems using the UI. See "
                    "https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Number"
                    "/MAX_SAFE_INTEGER"
                )
                warnings.warn(message, UserWarning, stacklevel=2)
        elif not isinstance(v, str):
            raise TypeError(f"Invalid type for id. Expected {int} or {str}; found:{type(v)}")
        return v

    @validator("prediction_agent", check_fields=False)
    def _check_prediction_agent(cls, v, values):
        """Triggers a warning when ONLY prediction agent is provided"""
        if v and values["prediction"] is None:
            warnings.warn(
                "You provided an `prediction_agent`, but no `prediction`. "
                "The `prediction_agent` will not be logged to the server."
            )
        return v

    @validator("annotation_agent", check_fields=False)
    def _check_annotation_agent(cls, v, values):
        """Triggers a warning when ONLY annotation agent is provided"""
        if v and values["annotation"] is None:
            warnings.warn(
                "You provided an `annotation_agent`, but no `annotation`. "
                "The `annotation_agent` will not be logged to the server."
            )
        return v

    @validator("event_timestamp", check_fields=False, always=True)
    def _nat_to_none_and_one_to_now(cls, v):
        """Converts pandas `NaT`s to `None`s and None´s to datetime.now()"""
        if v is pd.NaT:
            v = None

        v = v or datetime.datetime.now()

        return v

    @root_validator(skip_on_failure=True)
    def _check_and_update_status(cls, values):
        """Updates the status if an annotation is provided and no status is specified."""
        values["status"] = values.get("status") or ("Default" if values.get("annotation") is None else "Validated")

        return values

    class Config:
        # https://docs.pydantic.dev/usage/model_config/#options
        extra = "forbid"
        validate_assignment = True


class BulkResponse(BaseModel):
    """Summary response when logging records to the argilla server.

    Args:
        dataset: The dataset name.
        processed: Number of records in bulk.
        failed: Number of failed records.
    """

    dataset: str
    processed: int
    failed: Optional[int] = 0


class TokenAttributions(BaseModel):
    """Attribution of the token to the predicted label.

    In the argilla app this is only supported for ``TextClassificationRecord`` and the ``multi_label=False`` case.

    Args:
        token: The input token.
        attributions: A dictionary containing label-attribution pairs.
    """

    token: str
    attributions: Dict[str, float] = Field(default_factory=dict)


class TextClassificationRecord(_Validators):
    """Record for text classification

    Args:
        text:
            The input of the record. Provide either 'text' or 'inputs'.
        inputs:
            Various inputs of the record (see examples below).
            Provide either 'text' or 'inputs'.
        prediction:
            A list of tuples containing the predictions for the record.
            The first entry of the tuple is the predicted label, the second entry is its corresponding score.
        prediction_agent:
            Name of the prediction agent. By default, this is set to the hostname of your machine.
        annotation:
            A string or a list of strings (multilabel) corresponding to the annotation (gold label) for the record.
        annotation_agent:
            Name of the prediction agent. By default, this is set to the hostname of your machine.
        vectors:
            Vectors data mappings of the natural language text containing class attributes
        multi_label:
            Is the prediction/annotation for a multi label classification task? Defaults to `False`.
        explanation:
            A dictionary containing the attributions of each token to the prediction.
            The keys map the input of the record (see `inputs`) to the `TokenAttributions`.
        id:
            The id of the record. By default (`None`), we will generate a unique ID for you.
        metadata:
            Metadata for the record. Defaults to `{}`.
        status:
            The status of the record. Options: 'Default', 'Edited', 'Discarded', 'Validated'.
            If an annotation is provided, this defaults to 'Validated', otherwise 'Default'.
        event_timestamp:
            The timestamp for the creation of the record. Defaults to `datetime.datetime.now()`.
        metrics:
            READ ONLY! Metrics at record level provided by the server when using `rg.load`.
            This attribute will be ignored when using `rg.log`.
        search_keywords:
            READ ONLY! Relevant record keywords/terms for provided query when using `rg.load`.
            This attribute will be ignored when using `rg.log`.
    Examples:
        >>> # Single text input
        >>> import argilla as rg
        >>> record = rg.TextClassificationRecord(
        ...     text="My first argilla example",
        ...     prediction=[('eng', 0.9), ('esp', 0.1)],
        ...     vectors = {
        ...         "english_bert_vector": [1.2, 2.3, 3.1, 3.3]
        ...     }
        ... )
        >>>
        >>> # Various inputs
        >>> record = rg.TextClassificationRecord(
        ...     inputs={
        ...         "subject": "Has ganado 1 million!",
        ...         "body": "Por usar argilla te ha tocado este premio: <link>"
        ...     },
        ...     prediction=[('spam', 0.99), ('ham', 0.01)],
        ...     annotation="spam",
        ...     vectors = {
        ...                     "distilbert_uncased":  [1.13, 4.1, 6.3, 4.2, 9.1],
        ...                     "xlm_roberta_cased": [1.1, 2.1, 3.3, 4.2, 2.1],
        ...             }
        ...     )
    """

    text: Optional[str] = None
    inputs: Optional[Union[str, List[str], Dict[str, Union[str, List[str]]]]] = None

    prediction: Optional[List[Tuple[str, float]]] = None
    prediction_agent: Optional[str] = None

    annotation: Optional[Union[str, List[str]]] = None
    annotation_agent: Optional[str] = None

    vectors: Optional[Vectors] = None

    multi_label: bool = False
    explanation: Optional[Dict[str, List[TokenAttributions]]] = None

    id: Optional[Union[int, str]] = None
    metadata: Optional[Dict[str, Any]] = Field(default_factory=dict)
    status: Optional[str] = None
    event_timestamp: Optional[datetime.datetime] = None

    metrics: Optional[Dict[str, Any]] = None
    search_keywords: Optional[List[str]] = None

    @root_validator(skip_on_failure=True)
    def _check_text_and_inputs(cls, values):
        """Check if either text or inputs were provided. Copy text to inputs."""
        if isinstance(values.get("inputs"), str):
            warnings.warn(
                "In the future, the `inputs` argument of the `TextClassificationRecord`"
                " will not accept strings. Please use the `text` argument in that case."
                " Make sure to adapt your code accordingly.",
                category=FutureWarning,
            )

        if values.get("inputs") is not None and not isinstance(values["inputs"], dict):
            values["inputs"] = dict(text=values["inputs"])

        if (values.get("text") is None and values.get("inputs") is None) or (
            values.get("text") is not None
            and values.get("inputs") is not None
            and values["text"] != values["inputs"].get("text")
        ):
            raise ValueError("For a TextClassificationRecord you must provide either 'text' or 'inputs'")

        if values.get("text") is not None:
            values["inputs"] = dict(text=values["text"])
        elif len(values["inputs"]) == 1 and "text" in values["inputs"]:
            values["text"] = values["inputs"]["text"]

        return values

    def __setattr__(self, name: str, value: Any):
        """Make text and inputs immutable"""
        if name in ["text", "inputs"]:
            raise AttributeError(f"You cannot assign a new value to `{name}`")
        super().__setattr__(name, value)


class TokenClassificationRecord(_Validators):
    """Record for a token classification task

    Args:
        text:
            The input of the record
        tokens:
            The tokenized input of the record. We use this to guide the annotation process
            and to cross-check the spans of your `prediction`/`annotation`.
        prediction:
            A list of tuples containing the predictions for the record. The first entry of the tuple is the name of
            predicted entity, the second and third entry correspond to the start and stop character index of the entity.
            The fourth entry is optional and corresponds to the score of the entity (a float number between 0 and 1).
        prediction_agent:
            Name of the prediction agent. By default, this is set to the hostname of your machine.
        annotation:
            A list of tuples containing annotations (gold labels) for the record. The first entry of the tuple is the
            name of the entity, the second and third entry correspond to the start and stop char index of the entity.
        annotation_agent:
            Name of the prediction agent. By default, this is set to the hostname of your machine.
        vectors:
            Vector data mappings of the natural language text containing class attributes'
        id:
            The id of the record. By default (None), we will generate a unique ID for you.
        metadata:
            Metadata for the record. Defaults to `{}`.
        status:
            The status of the record. Options: 'Default', 'Edited', 'Discarded', 'Validated'.
            If an annotation is provided, this defaults to 'Validated', otherwise 'Default'.
        event_timestamp:
            The timestamp for the creation of the record. Defaults to `datetime.datetime.now()`.
        metrics:
            READ ONLY! Metrics at record level provided by the server when using `rg.load`.
            This attribute will be ignored when using `rg.log`.
        search_keywords:
            READ ONLY! Relevant record keywords/terms for provided query when using `rg.load`.
            This attribute will be ignored when using `rg.log`.
    Examples:
        >>> import argilla as rg
        >>> record = rg.TokenClassificationRecord(
        ...     text = "Michael is a professor at Harvard",
        ...     tokens = ["Michael", "is", "a", "professor", "at", "Harvard"],
        ...     prediction = [('NAME', 0, 7), ('LOC', 26, 33)],
        ...     vectors = {
        ...            "bert_base_uncased": [3.2, 4.5, 5.6, 8.9]
        ...          }
        ... )
    """

    text: Optional[str] = Field(None, min_length=1)
    tokens: Optional[Union[List[str], Tuple[str, ...]]] = None

    prediction: Optional[List[Union[Tuple[str, int, int], Tuple[str, int, int, Optional[float]]]]] = None
    prediction_agent: Optional[str] = None
    annotation: Optional[List[Tuple[str, int, int]]] = None
    annotation_agent: Optional[str] = None
    vectors: Optional[Vectors] = None

    id: Optional[Union[int, str]] = None
    metadata: Optional[Dict[str, Any]] = Field(default_factory=dict)
    status: Optional[str] = None
    event_timestamp: Optional[datetime.datetime] = None

    metrics: Optional[Dict[str, Any]] = None
    search_keywords: Optional[List[str]] = None
    _span_utils: SpanUtils = PrivateAttr()

    def __init__(
        self,
        text: str = None,
        tokens: List[str] = None,
        tags: Optional[List[str]] = None,
        **data,
    ):
        if text is None and tokens is None:
            raise AssertionError("Missing fields: At least one of `text` or `tokens` argument must be provided!")

        if (data.get("annotation") or data.get("prediction")) and text is None:
            raise AssertionError("Missing field `text`: " "char level spans must be provided with a raw text sentence")

        if text is None:
            text = " ".join(tokens)

        super().__init__(text=text, tokens=tokens, **data)

        self._span_utils = SpanUtils(self.text, self.tokens)

        if self.annotation:
            self.annotation = self._validate_spans(self.annotation)
        if self.prediction:
            self.prediction = self._validate_spans(self.prediction)

        if self.annotation and tags:
            _LOGGER.warning("Annotation already provided, `tags` won't be used")
        elif tags:
            self.annotation = self._span_utils.from_tags(tags)

    def __setattr__(self, name: str, value: Any):
        """Make text and tokens immutable"""
        if name in ["text", "tokens"]:
            raise AttributeError(f"You cannot assign a new value to `{name}`")
        super().__setattr__(name, value)

    def _validate_spans(self, spans: List[Tuple[str, int, int]]) -> List[Tuple[str, int, int]]:
        """Validates the entity spans with respect to the tokens.

        If necessary, also performs an automatic correction of the spans.

        Args:
            spans: The entity spans to validate.

        Returns:
            The optionally corrected spans.

        Raises:
            ValidationError: If spans are not valid or misaligned.
        """
        try:
            self._span_utils.validate(spans)
        except ValueError:
            spans = self._span_utils.correct(spans)
            self._span_utils.validate(spans)

        return spans

    @validator("tokens", pre=True)
    def _normalize_tokens(cls, value):
        if isinstance(value, list):
            value = tuple(value)

        assert len(value) > 0, "At least one token should be provided"
        return value

    @validator("prediction")
    def _add_default_score(
        cls,
        prediction: Optional[List[Union[Tuple[str, int, int], Tuple[str, int, int, Optional[float]]]]],
    ):
        """Adds the default score to the predictions if it is missing"""
        if prediction is None:
            return prediction
        return [
            (pred[0], pred[1], pred[2], 0.0) if len(pred) == 3 else (pred[0], pred[1], pred[2], pred[3] or 0.0)
            for pred in prediction
        ]

    @validator("text")
    def _check_if_empty_after_strip(cls, text: str):
        assert text.strip(), "The provided `text` contains only whitespaces."
        return text

    @property
    def __chars2tokens__(self) -> Dict[int, int]:
        """DEPRECATED, please use the ``argilla.utils.span_utils.SpanUtils.chars_to_token_idx`` attribute."""
        warnings.warn(
            "The `__chars2tokens__` attribute is deprecated and will be removed in a"
            " future version. Please use the"
            " `argilla.utils.span_utils.SpanUtils.char_to_token_idx` attribute"
            " instead.",
            FutureWarning,
        )
        return self._span_utils.char_to_token_idx

    @property
    def __tokens2chars__(self) -> Dict[int, Tuple[int, int]]:
        """DEPRECATED, please use the ``argilla.utils.span_utils.SpanUtils.chars_to_token_idx`` attribute."""
        warnings.warn(
            "The `__tokens2chars__` attribute is deprecated and will be removed in a"
            " future version. Please use the"
            " `argilla.utils.span_utils.SpanUtils.token_to_char_idx` attribute"
            " instead.",
            FutureWarning,
        )
        return self._span_utils.token_to_char_idx

    def char_id2token_id(self, char_idx: int) -> Optional[int]:
        """DEPRECATED, please use the ``argilla.utisl.span_utils.SpanUtils.char_to_token_idx`` dict instead."""
        warnings.warn(
            "The `char_id2token_id` method is deprecated and will be removed in a"
            " future version. Please use the"
            " `argilla.utils.span_utils.SpanUtils.char_to_token_idx` dict instead.",
            FutureWarning,
        )
        return self._span_utils.char_to_token_idx.get(char_idx)

    def token_span(self, token_idx: int) -> Tuple[int, int]:
        """DEPRECATED, please use the ``argilla.utisl.span_utils.SpanUtils.token_to_char_idx`` dict instead."""
        warnings.warn(
            "The `token_span` method is deprecated and will be removed in a future"
            " version. Please use the"
            " `argilla.utils.span_utils.SpanUtils.token_to_char_idx` dict instead.",
            FutureWarning,
        )
        if token_idx not in self._span_utils.token_to_char_idx:
            raise IndexError(f"Token id {token_idx} out of bounds")
        return self._span_utils.token_to_char_idx[token_idx]

    def spans2iob(self, spans: Optional[List[Tuple[str, int, int]]] = None) -> Optional[List[str]]:
        """DEPRECATED, please use the ``argilla.utils.SpanUtils.to_tags()`` method."""
        warnings.warn(
            "'spans2iob' is deprecated and will be removed in a future version. Please"
            " use the `argilla.utils.SpanUtils.to_tags()` method instead, and adapt"
            " your code accordingly.",
            FutureWarning,
        )

        if spans is None:
            return None
        return self._span_utils.to_tags(spans)


class Text2TextRecord(_Validators):
    """Record for a text to text task

    Args:
        text:
            The input of the record
        prediction:
            A list of strings or tuples containing predictions for the input text.
            If tuples, the first entry is the predicted text, the second entry is its corresponding score.
        prediction_agent:
            Name of the prediction agent. By default, this is set to the hostname of your machine.
        annotation:
            A string representing the expected output text for the given input text.
        annotation_agent:
            Name of the prediction agent. By default, this is set to the hostname of your machine.
        vectors:
            Embedding data mappings of the natural language text containing class attributes'
        id:
            The id of the record. By default (None), we will generate a unique ID for you.
        metadata:
            Metadata for the record. Defaults to `{}`.
        status:
            The status of the record. Options: 'Default', 'Edited', 'Discarded', 'Validated'.
            If an annotation is provided, this defaults to 'Validated', otherwise 'Default'.
        event_timestamp:
            The timestamp for the creation of the record. Defaults to `datetime.datetime.now()`.
        metrics:
            READ ONLY! Metrics at record level provided by the server when using `rg.load`.
            This attribute will be ignored when using `rg.log`.
        search_keywords:
            READ ONLY! Relevant record keywords/terms for provided query when using `rg.load`.
            This attribute will be ignored when using `rg.log`.

    Examples:
        >>> import argilla as rg
        >>> record = rg.Text2TextRecord(
        ...     text="My name is Sarah and I love my dog.",
        ...     prediction=["Je m'appelle Sarah et j'aime mon chien."],
        ...     vectors = {
        ...         "bert_base_uncased": [1.2, 2.3, 3.4, 5.2, 6.5],
        ...         "xlm_multilingual_uncased": [2.2, 5.3, 5.4, 3.2, 2.5]
        ...     }
        ... )
    """

    text: str

    prediction: Optional[List[Union[str, Tuple[str, float]]]] = None
    prediction_agent: Optional[str] = None
    annotation: Optional[str] = None
    annotation_agent: Optional[str] = None
    vectors: Optional[Vectors] = None

    id: Optional[Union[int, str]] = None
    metadata: Optional[Dict[str, Any]] = Field(default_factory=dict)
    status: Optional[str] = None
    event_timestamp: Optional[datetime.datetime] = None

    metrics: Optional[Dict[str, Any]] = None
    search_keywords: Optional[List[str]] = None

    @validator("prediction")
    def prediction_as_tuples(cls, prediction: Optional[List[Union[str, Tuple[str, float]]]]):
        """Preprocess the predictions and wraps them in a tuple if needed"""
        if prediction is None:
            return prediction
        return [(pred, 1.0) if isinstance(pred, str) else pred for pred in prediction]


@deprecated("Use Text2TextRecord instead.")
class TextGenerationRecord(Text2TextRecord):  # TODO Remove TextGenerationRecord
    pass


Record = Union[
    TextClassificationRecord,
    TokenClassificationRecord,
    Text2TextRecord,
    TextGenerationRecord,
]



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Dict


def empty_visualization(message: str = "No data found"):
    import plotly.graph_objects as go

    fig = go.Figure()
    fig.update_layout(
        xaxis={"visible": False},
        yaxis={"visible": False},
        annotations=[{"text": message, "showarrow": False, "font": {"size": 30}}],
    )
    return fig


def bar(data: dict, title: str = "Bar", x_legend: str = "", y_legend: str = ""):
    import plotly.graph_objects as go

    if not data:
        return empty_visualization()

    keys, values = zip(*data.items())
    keys = [key.encode("unicode-escape").decode() if isinstance(key, str) else key for key in keys]
    fig = go.Figure(data=go.Bar(y=values, x=keys))
    fig.update_layout(
        title=title,
        xaxis_title=x_legend,
        yaxis_title=y_legend,
    )
    return fig


def stacked_bar(x: list, y_s: Dict[str, list], title: str = "Bar", x_legend: str = "", y_legend: str = ""):
    import plotly.graph_objects as go

    if not x or not y_s:
        return empty_visualization()

    data = [go.Bar(name=name, x=x, y=y_values) for name, y_values in y_s.items()]

    fig = go.Figure(data=data)
    fig.update_layout(
        title=title,
        xaxis_title=x_legend,
        yaxis_title=y_legend,
        barmode="stack",
    )

    return fig


def histogram(data: dict, title: str = "Bar", x_legend: str = "", y_legend: str = ""):
    if not data:
        return empty_visualization()

    data = {float(k): v for k, v in data.items()}
    return bar(data, title, x_legend, y_legend)


def tree_map(labels, parents, values, title: str = "Tree"):
    import plotly.graph_objects as go

    if not values:  # TODO: review condition
        return empty_visualization()

    fig = go.Figure(
        go.Treemap(
            labels=labels,
            parents=parents,
            values=values,
        )
    )
    fig.update_layout(title=title, margin=dict(t=50, l=0, r=0, b=0))

    return fig


def multilevel_pie(labels, parents, values, title: str = "Pie"):
    import plotly.graph_objects as go

    fig = go.Figure(
        go.Sunburst(
            labels=labels,
            parents=parents,
            values=values,
        )
    )
    fig.update_layout(title=title, margin=dict(t=50, l=0, r=0, b=0))

    return fig


def f1(data: Dict[str, float], title: str):
    from plotly.subplots import make_subplots

    fig = make_subplots(
        rows=1,
        cols=3,
        subplot_titles=[
            "macro average",
            "micro average",
            "per label",
        ],
    )

    fig.add_bar(
        x=[k.split("_")[0] for k, v in data.items() if "macro" in k],
        y=[v for k, v in data.items() if "macro" in k],
        row=1,
        col=1,
    )
    fig.add_bar(
        x=[k.split("_")[0] for k, v in data.items() if "micro" in k],
        y=[v for k, v in data.items() if "micro" in k],
        row=1,
        col=2,
    )
    per_label = {k: v for k, v in data.items() if all(key not in k for key in ["macro", "micro", "support"])}

    fig.add_bar(
        x=[k for k, v in per_label.items()],
        y=[v for k, v in per_label.items()],
        row=1,
        col=3,
    )
    fig.update_layout(showlegend=False, title_text=title)

    return fig



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Callable, Dict, Hashable, List, Tuple, Union

import pandas as pd

from argilla.client.feedback.schemas.remote.shared import RemoteSchema
from argilla.pydantic_v1 import BaseModel

if TYPE_CHECKING:
    from argilla.client.feedback.dataset import FeedbackDataset


# Type aliases
Responses = List[Union[float, int, str]]
Suggestions = Responses
# Expected format for the nltk's AnnotationTask
FormattedResponses = List[Tuple[Any, Hashable, Hashable]]


class MetricResultBase(BaseModel):
    """Base class for the result of a metric."""

    metric_name: str
    count: int


class AgreementMetricResult(MetricResultBase):
    """Container for the result of an agreement metric.

    It contains two fields, `metric_name` and `result` with the value of the metric.
    """

    result: float


class ModelMetricResult(MetricResultBase):
    """Container for the result of an annotator metric.

    It contains two fields, `metric_name` and `result` with the value of the metric.
    """

    result: Union[float, Dict[str, float], pd.DataFrame, Dict[str, pd.DataFrame]]

    class Config:
        arbitrary_types_allowed = True


class AnnotatorMetricBase(ABC):
    """Base class for Annotator metrics."""

    def __init__(self, responses: Responses = None, suggestions: Suggestions = None) -> None:
        """
        Args:
            responses: Responses given by the user.
                Depending on the type of question it can be a list of strings, or integers.
            suggestions: Suggestions offered for the annotators.
                Same format as `responses`.
        """
        self._responses = responses
        self._suggestions = suggestions

    def compute(self, **kwargs):
        responses, suggestions = self._pre_process(self._responses, self._suggestions)
        return self._compute(responses, suggestions, **kwargs)

    def _pre_process(self, responses: Responses, suggestions: Suggestions) -> Any:
        """Optional data preprocessing. By default it just passes the data to the _compute method.

        Args:
            responses: Responses given by the user.
            suggestions: Suggestions offered for the annotators.

        Returns:
            data: tuple with the preprocessed data.
        """
        return responses, suggestions

    @abstractmethod
    def _compute(self, responses: Responses, suggestions: Suggestions, **kwargs):
        """Abstract method where the computation is done.

        Args:
            responses: Responses given by the user, as expected by the given metric.
            suggestions: Suggestions offered for the annotators, as expected by the given metric.
        """
        pass


class AnnotationTaskMetricBase(ABC):
    """Base class for Agreement metrics."""

    def __init__(self, annotated_dataset: FormattedResponses = None, distance_function: Callable = lambda x: x) -> None:
        """
        Args:
            annotated_dataset: Annotated dataset as expected by the metric.
            distance_function: Distance function to use for the metric.
                Depending on the type of data we need a function to compute the distance.
                For example for binary data we can use the binary distance function,
                while RatingQuestion works with an interval distance as we are dealing with
                numeric values.
        """
        self._dataset = annotated_dataset
        self._distance_function = distance_function

    def compute(self) -> float:
        """General method to obtain the metric.

        Args:
            kwargs: Optional arguments that could be passed to the metric.

        Returns:
            metric: Metric result that will be stored in the `AgreementMetricResult`.
        """
        data = self._pre_process(self._dataset)
        return self._compute(data)

    def _pre_process(self, data: FormattedResponses) -> Any:
        """Optional data preprocessing. By default it just passes the data to the _compute method.

        Args:
            data: annotated dataset.
            kwargs: optional arguments to be passed to the metric.

        Returns:
            data: dataset prepared for the _compute method.
        """
        return data

    @abstractmethod
    def _compute(self, data: FormattedResponses):
        """Abstract method where the computation is done.

        Args:
            data: Data as expected for the given metric.
        """
        pass


class MetricBase:
    _metrics_per_question: Dict[str, Callable] = {}

    def __init__(self, dataset: "FeedbackDataset", question_name: str, responses_vs_suggestions: bool = True) -> None:
        """Initializes a `AgreementMetric` object to compute agreement metrics on
        a `FeedbackDataset` for a given question.

        Args:
            dataset: FeedbackDataset to compute the metrics.
            question_name: Name of the question for which we want to analyse the agreement.
            responses_vs_suggestions: Whether to compare the responses vs the suggestions, or the
                other way around. Defaults to True (the metrics will be compared assuming the
                responses are the ground truth and the suggestions are the predictions).

        Raises:
            NotImplementedError: If the question type is not supported.
        """
        self._dataset = dataset
        self._question_name = question_name
        self._question_type = type(self._dataset.question_by_name(question_name))

        # Check to assume the remote questions behave just like local ones
        if issubclass(self._question_type, RemoteSchema):
            self._question_type = type(self._dataset.question_by_name(question_name).to_local())

        if allowed_metrics := self._metrics_per_question.get(self._question_type):
            self._allowed_metrics = allowed_metrics
        else:
            raise NotImplementedError(f"No metrics are defined currently for {self._question_type.__name__}")
        self._responses_vs_suggestions = responses_vs_suggestions

    def __repr__(self) -> str:
        return type(self).__name__ + f"(question_name={self._question_name})"

    @property
    def allowed_metrics(self) -> List[str]:
        """Available metrics for the given question."""
        return list(self._allowed_metrics)

    def _check_metrics(self, metric_names: Union[str, List[str]]) -> List[str]:
        if isinstance(metric_names, str):
            metric_names = [metric_names]

        if any([metric not in self._allowed_metrics for metric in metric_names]):
            raise ValueError(
                f"Metrics allowed for question {self._question_name}: {list(self._allowed_metrics.keys())}"
            )
        return metric_names

    def _get_metric_classes(self, metric_names: Union[str, List[str]]) -> List[Tuple[str, Callable]]:
        return [(metric_name, self._allowed_metrics[metric_name]) for metric_name in metric_names]

    def _prepare_responses_and_suggestions(
        self, responses: Responses, suggestions: Responses
    ) -> Union[Tuple[Responses, Suggestions], Tuple[Suggestions, Responses]]:
        """Helper function to determine the order in which the responses and suggestions should be passed to the metric,
        to avoid duplicating code in the metrics.

        Args:
            responses: Responses
            suggestions: Responses

        Returns:
            Union[Tuple[Responses, Suggestions], Tuple[Suggestions, Responses]]
        """
        if self._responses_vs_suggestions:
            return responses, suggestions
        else:
            return suggestions, responses
